import logging
import math

import numpy as np

from flaml.tune import Trial
from flaml.tune.scheduler import TrialScheduler

logger = logging.getLogger(__name__)


class OnlineTrialRunner:
    """Class for the OnlineTrialRunner."""

    # ************NOTE about the status of a trial***************
    # Trial.PENDING: All trials are set to be pending when frist added into the OnlineTrialRunner until
    #     it is selected to run. By this definition, a trial with status Trial.PENDING is a challenger
    #     trial added to the OnlineTrialRunner but never been selected to run.
    #     It denotes the starting of trial's lifespan in the OnlineTrialRunner.
    # Trial.RUNNING: It indicates that this trial is one of the concurrently running trials.
    #     The max number of Trial.RUNNING trials is running_budget.
    #     The status of a trial will be set to Trial.RUNNING the next time it selected to run.
    #     A trial's status may have the following change:
    #     Trial.PENDING -> Trial.RUNNING
    #     Trial.PAUSED - > Trial.RUNNING
    # Trial.PAUSED: The status of a trial is set to Trial.PAUSED once it is removed from the running trials.
    #     Trial.RUNNING - > Trial.PAUSED
    # Trial.TERMINATED: set the status of a trial to Trial.TERMINATED when you never want to select it.
    #     It denotes the real end of a trial's lifespan.
    # Status change routine of a trial:
    # Trial.PENDING -> (Trial.RUNNING -> Trial.PAUSED -> Trial.RUNNING -> ...) -> Trial.TERMINATED(optional)

    RANDOM_SEED = 123456
    WARMSTART_NUM = 100

    def __init__(
        self, max_live_model_num: int, searcher=None, scheduler=None, champion_test_policy="loss_ucb", **kwargs
    ):
        """Constructor.

        Args:
            max_live_model_num: The maximum number of 'live'/running models allowed.
            searcher: A class for generating Trial objects progressively.
                The ConfigOracle is implemented in the searcher.
            scheduler: A class for managing the 'live' trials and allocating the
                resources for the trials.
            champion_test_policy: A string to specify what test policy to test for
                champion. Currently can choose from ['loss_ucb', 'loss_avg', 'loss_lcb', None].
        """
        # ************A NOTE about the input searcher and scheduler******
        # Required methods of the searcher:
        # - next_trial()
        #     Generate the next trial to add.
        # - set_search_properties(metric: Optional[str], mode: Optional[str],
        #                         config: Optional[dict], setting: Optional[dict])
        #     Generate new challengers based on the current champion and update the challenger list
        # - on_trial_result(trial_id: str, result: Dict)
        #     Reprot results to the scheduler.
        # Required methods of the scheduler:
        # - on_trial_add(trial_runner, trial: Trial)
        #     It adds candidate trials to the scheduler. It is called inside of the add_trial
        #     function in the TrialRunner.
        # - on_trial_remove(trial_runner, trial: Trial)
        #     Remove terminated trials from the scheduler.
        # - on_trial_result(trial_runner, trial: Trial, result: Dict)
        #     Reprot results to the scheduler.
        # - choose_trial_to_run(trial_runner) -> Optional[Trial]
        # Among them, on_trial_result and choose_trial_to_run are the most important methods
        # *****************************************************************
        # OnlineTrialRunner setting
        self._searcher = searcher
        self._scheduler = scheduler
        self._champion_test_policy = champion_test_policy
        self._max_live_model_num = max_live_model_num
        self._remove_worse = kwargs.get("remove_worse", True)
        self._bound_trial_num = kwargs.get("bound_trial_num", False)
        self._no_model_persistence = True

        # stores all the trials added to the OnlineTrialRunner
        # i.e., include the champion and all the challengers
        self._trials = []
        self._champion_trial = None
        self._best_challenger_trial = None
        self._first_challenger_pool_size = None
        self._random_state = np.random.RandomState(self.RANDOM_SEED)
        self._running_trials = set()

        # initially schedule up to max_live_model_num of live models and
        # set the first trial as the champion (which is done inside self.step())
        self._total_steps = 0
        logger.info("init step %s", self._max_live_model_num)
        # TODO: add more comments
        self.step()
        assert self._champion_trial is not None

    @property
    def champion_trial(self) -> Trial:
        """The champion trial."""
        return self._champion_trial

    @property
    def running_trials(self):
        """The running/'live' trials."""
        return self._running_trials

    def step(self, data_sample=None, prediction_trial_tuple=None):
        """Schedule one trial to run each time it is called.

        Args:
            data_sample: One data example.
            prediction_trial_tuple: A list of information containing
                (prediction_made, prediction_trial).
        """
        # TODO: Will remove prediction_trial_tuple.
        # NOTE: This function consists of the following several parts:
        # * Update model:
        # 0. Update running trials using observations received.
        # * Tests for Champion:
        # 1. Test for champion (BetterThan test, and WorseThan test)
        #     1.1 BetterThan test
        #     1.2 WorseThan test: a trial may be removed if WroseThan test is triggered
        # * Online Scheduling:
        # 2. Report results to the searcher and scheduler (the scheduler will return a decision about
        #     the status of the running trials).
        # 3. Pause or stop a trial according to the scheduler's decision.
        # Add a trial into the OnlineTrialRunner if there are opening slots.

        # ***********Update running trials with observation*******************
        if data_sample is not None:
            self._total_steps += 1
            prediction_made, prediction_trial = (
                prediction_trial_tuple[0],
                prediction_trial_tuple[1],
            )
            # assert prediction_trial.status == Trial.RUNNING
            trials_to_pause = []
            for trial in list(self._running_trials):
                if trial != prediction_trial:
                    y_predicted = trial.predict(data_sample)
                else:
                    y_predicted = prediction_made
                trial.train_eval_model_online(data_sample, y_predicted)
                logger.debug(
                    "running trial at iter %s %s %s %s %s %s",
                    self._total_steps,
                    trial.trial_id,
                    trial.result.loss_avg,
                    trial.result.loss_cb,
                    trial.result.resource_used,
                    trial.resource_lease,
                )
                # report result to the searcher
                self._searcher.on_trial_result(trial.trial_id, trial.result)
                # report result to the scheduler and the scheduler makes a decision about
                # the running status of the trial
                decision = self._scheduler.on_trial_result(self, trial, trial.result)
                # set the status of the trial according to the decision made by the scheduler
                logger.debug(
                    "trial decision %s %s at step %s",
                    decision,
                    trial.trial_id,
                    self._total_steps,
                )
                if decision == TrialScheduler.STOP:
                    self.stop_trial(trial)
                elif decision == TrialScheduler.PAUSE:
                    trials_to_pause.append(trial)
                else:
                    self.run_trial(trial)
            # ***********Statistical test of champion*************************************
            self._champion_test()
            # Pause the trial after the tests because the tests involves the reset of the trial's result
            for trial in trials_to_pause:
                self.pause_trial(trial)
        # ***********Add and schedule new trials to run if there are opening slots****
        # Add trial if needed: add challengers into consideration through _add_trial_from_searcher()
        # if there are available slots
        for _ in range(self._max_live_model_num - len(self._running_trials)):
            self._add_trial_from_searcher()
        # Scheduling: schedule up to max_live_model_num number of trials to run
        # (set the status as Trial.RUNNING)
        while self._max_live_model_num > len(self._running_trials):
            trial_to_run = self._scheduler.choose_trial_to_run(self)
            if trial_to_run is not None:
                self.run_trial(trial_to_run)
            else:
                break

    def get_top_running_trials(self, top_ratio=None, top_metric="ucb") -> list:
        """Get a list of trial ids, whose performance is among the top running trials."""
        running_valid_trials = [trial for trial in self._running_trials if trial.result is not None]
        if not running_valid_trials:
            return
        if top_ratio is None:
            top_number = 0
        elif isinstance(top_ratio, float):
            top_number = math.ceil(len(running_valid_trials) * top_ratio)
        elif isinstance(top_ratio, str) and "best" in top_ratio:
            top_number = 1
        else:
            raise NotImplementedError

        if "ucb" in top_metric:
            test_attribute = "loss_ucb"
        elif "avg" in top_metric:
            test_attribute = "loss_avg"
        elif "lcb" in top_metric:
            test_attribute = "loss_lcb"
        else:
            raise NotImplementedError
        top_running_valid_trials = []
        logger.info("Running trial ids %s", [trial.trial_id for trial in running_valid_trials])
        self._random_state.shuffle(running_valid_trials)
        results = [trial.result.get_score(test_attribute) for trial in running_valid_trials]
        # sorted result (small to large) index
        sorted_index = np.argsort(np.array(results))
        for i in range(min(top_number, len(running_valid_trials))):
            top_running_valid_trials.append(running_valid_trials[sorted_index[i]])
        logger.info("Top running ids %s", [trial.trial_id for trial in top_running_valid_trials])
        return top_running_valid_trials

    def _add_trial_from_searcher(self):
        """Add a new trial to this TrialRunner.

        NOTE:
            The new trial is acquired from the input search algorithm, i.e. self._searcher.
            A 'new' trial means the trial is not in self._trial.
        """
        # (optionally) upper bound the number of trials in the OnlineTrialRunner
        if self._bound_trial_num and self._first_challenger_pool_size is not None:
            active_trial_size = len([t for t in self._trials if t.status != Trial.TERMINATED])
            trial_num_upper_bound = (
                int(round((np.log10(self._total_steps) + 1) * self._first_challenger_pool_size))
                if self._first_challenger_pool_size
                else np.inf
            )
            if active_trial_size > trial_num_upper_bound:
                logger.info(
                    "Not adding new trials: %s exceeds trial limit %s.",
                    active_trial_size,
                    trial_num_upper_bound,
                )
                return None

        # output one trial from the trial pool (new challenger pool) maintained in the searcher
        # Assumption on the searcher: when all frontiers (i.e., all the challengers generated
        # based on the current champion) of the current champion are added, calling next_trial()
        # will return None
        trial = self._searcher.next_trial()
        if trial is not None:
            self.add_trial(trial)  # dup checked in add_trial
            # the champion_trial is initially None, so we need to set it up the first time
            # a valid trial is added.
            # Assumption on self._searcher: the first trial generated is the champion trial
            if self._champion_trial is None:
                logger.info("Initial set up of the champion trial %s", trial.config)
                self._set_champion(trial)
        else:
            self._all_new_challengers_added = True
            if self._first_challenger_pool_size is None:
                self._first_challenger_pool_size = len(self._trials)

    def _champion_test(self):
        """Perform tests again the latest champion, including bette_than tests and worse_than tests"""
        # for BetterThan test, we only need to compare the best challenger with the champion
        self._get_best_challenger()
        if self._best_challenger_trial is not None:
            assert self._best_challenger_trial.trial_id != self._champion_trial.trial_id
            # test whether a new champion is found and set the trial properties accordingly
            is_new_champion_found = self._better_than_champion_test(self._best_challenger_trial)
            if is_new_champion_found:
                self._set_champion(new_champion_trial=self._best_challenger_trial)

        # performs _worse_than_champion_test, which is an optional component in ChaCha
        if self._remove_worse:
            to_stop = []
            for trial_to_test in self._trials:
                if trial_to_test.status != Trial.TERMINATED:
                    worse_than_champion = self._worse_than_champion_test(
                        self._champion_trial, trial_to_test, self.WARMSTART_NUM
                    )
                    if worse_than_champion:
                        to_stop.append(trial_to_test)
            # we want to ensure there are at least #max_live_model_num of challengers remaining
            max_to_stop_num = len([t for t in self._trials if t.status != Trial.TERMINATED]) - self._max_live_model_num
            for i in range(min(max_to_stop_num, len(to_stop))):
                self.stop_trial(to_stop[i])

    def _get_best_challenger(self):
        """Get the 'best' (in terms of the champion_test_policy) challenger under consideration."""
        if self._champion_test_policy is None:
            return
        if "ucb" in self._champion_test_policy:
            test_attribute = "loss_ucb"
        elif "avg" in self._champion_test_policy:
            test_attribute = "loss_avg"
        else:
            raise NotImplementedError
        active_trials = [
            trial
            for trial in self._trials
            if (
                trial.status != Trial.TERMINATED
                and trial.trial_id != self._champion_trial.trial_id
                and trial.result is not None
            )
        ]
        if active_trials:
            self._random_state.shuffle(active_trials)
            results = [trial.result.get_score(test_attribute) for trial in active_trials]
            best_index = np.argmin(results)
            self._best_challenger_trial = active_trials[best_index]

    def _set_champion(self, new_champion_trial):
        """Set the status of the existing trials once a new champion is found."""
        assert new_champion_trial is not None
        is_init_update = False
        if self._champion_trial is None:
            is_init_update = True
        self.run_trial(new_champion_trial)
        # set the checked_under_current_champion status of the trials
        for trial in self._trials:
            if trial.trial_id == new_champion_trial.trial_id:
                trial.set_checked_under_current_champion(True)
            else:
                trial.set_checked_under_current_champion(False)
        self._champion_trial = new_champion_trial
        self._all_new_challengers_added = False
        logger.info("Set the champion as %s", self._champion_trial.trial_id)
        if not is_init_update:
            self._champion_update_times += 1
            # calling set_search_properties of searcher will trigger
            # new challenger generation. we do not do this for init champion
            # as this step is already done when first constructing the searcher
            self._searcher.set_search_properties(setting={self._searcher.CHAMPION_TRIAL_NAME: self._champion_trial})
        else:
            self._champion_update_times = 0

    def get_trials(self) -> list:
        """Return the list of trials managed by this TrialRunner."""
        return self._trials

    def add_trial(self, new_trial):
        """Add a new trial to this TrialRunner.
        Trials may be added at any time.

        Args:
            new_trial (Trial): Trial to queue.
        """
        # Only add the new trial when it does not exist (according to the trial_id, which is
        # the signature of the trail) in self._trials.
        for trial in self._trials:
            if trial.trial_id == new_trial.trial_id:
                trial.set_checked_under_current_champion(True)
                return
        logger.info(
            "adding trial at iter %s, %s %s",
            self._total_steps,
            new_trial.trial_id,
            len(self._trials),
        )
        self._trials.append(new_trial)
        self._scheduler.on_trial_add(self, new_trial)

    def stop_trial(self, trial):
        """Stop a trial: set the status of a trial to be
        Trial.TERMINATED and perform other subsequent operations.
        """
        if trial.status in [Trial.ERROR, Trial.TERMINATED]:
            return
        else:
            logger.info(
                "Terminating trial %s, with trial result %s",
                trial.trial_id,
                trial.result,
            )
            trial.set_status(Trial.TERMINATED)
            # clean up model and result
            trial.clean_up_model()
            self._scheduler.on_trial_remove(self, trial)
            self._searcher.on_trial_complete(trial.trial_id)
            self._running_trials.remove(trial)

    def pause_trial(self, trial):
        """Pause a trial: set the status of a trial to be Trial.PAUSED
        and perform other subsequent operations.
        """
        if trial.status in [Trial.ERROR, Trial.TERMINATED]:
            return
        else:
            logger.info(
                "Pausing trial %s, with trial loss_avg: %s, loss_cb: %s, loss_ucb: %s,\
                        resource_lease: %s",
                trial.trial_id,
                trial.result.loss_avg,
                trial.result.loss_cb,
                trial.result.loss_avg + trial.result.loss_cb,
                trial.resource_lease,
            )
            trial.set_status(Trial.PAUSED)
            # clean up model and result if no model persistence
            if self._no_model_persistence:
                trial.clean_up_model()
            self._running_trials.remove(trial)

    def run_trial(self, trial):
        """Run a trial: set the status of a trial to be Trial.RUNNING
        and perform other subsequent operations.
        """
        if trial.status in [Trial.ERROR, Trial.TERMINATED]:
            return
        else:
            trial.set_status(Trial.RUNNING)
            self._running_trials.add(trial)

    def _better_than_champion_test(self, trial_to_test):
        """Test whether there is a config in the existing trials that
        is better than the current champion config.

        Returns:
            A bool indicating whether a new champion is found.
        """
        if trial_to_test.result is not None and self._champion_trial.result is not None:
            if "ucb" in self._champion_test_policy:
                return self._test_lcb_ucb(self._champion_trial, trial_to_test, self.WARMSTART_NUM)
            elif "avg" in self._champion_test_policy:
                return self._test_avg_loss(self._champion_trial, trial_to_test, self.WARMSTART_NUM)
            elif "martingale" in self._champion_test_policy:
                return self._test_martingale(self._champion_trial, trial_to_test)
            else:
                raise NotImplementedError
        else:
            return False

    @staticmethod
    def _worse_than_champion_test(champion_trial, trial, warmstart_num=1) -> bool:
        """Test whether the input trial is worse than the champion_trial"""
        if trial.result is not None and trial.result.resource_used >= warmstart_num:
            if trial.result.loss_lcb > champion_trial.result.loss_ucb:
                logger.info(
                    "=========trial %s is worse than champion %s=====",
                    trial.trial_id,
                    champion_trial.trial_id,
                )
                logger.info("trial %s %s %s", trial.config, trial.result, trial.resource_lease)
                logger.info(
                    "trial loss_avg:%s, trial loss_cb %s",
                    trial.result.loss_avg,
                    trial.result.loss_cb,
                )
                logger.info(
                    "champion loss_avg:%s, champion loss_cb %s",
                    champion_trial.result.loss_avg,
                    champion_trial.result.loss_cb,
                )
                logger.info("champion %s", champion_trial.config)
                logger.info(
                    "trial loss_avg_recent:%s, trial loss_cb %s",
                    trial.result.loss_avg_recent,
                    trial.result.loss_cb,
                )
                logger.info(
                    "champion loss_avg_recent:%s, champion loss_cb %s",
                    champion_trial.result.loss_avg_recent,
                    champion_trial.result.loss_cb,
                )
                return True
        return False

    @staticmethod
    def _test_lcb_ucb(champion_trial, trial, warmstart_num=1) -> bool:
        """Comare the challenger(i.e., trial)'s loss upper bound with
        champion_trial's loss lower bound - cb
        """
        assert trial.trial_id != champion_trial.trial_id
        if trial.result.resource_used >= warmstart_num:
            if trial.result.loss_ucb < champion_trial.result.loss_lcb - champion_trial.result.loss_cb:
                logger.info("======new champion condition satisfied: using lcb vs ucb=====")
                logger.info(
                    "new champion trial %s %s %s",
                    trial.trial_id,
                    trial.result.resource_used,
                    trial.resource_lease,
                )
                logger.info(
                    "new champion trial loss_avg:%s, trial loss_cb %s",
                    trial.result.loss_avg,
                    trial.result.loss_cb,
                )
                logger.info(
                    "old champion trial %s %s %s",
                    champion_trial.trial_id,
                    champion_trial.result.resource_used,
                    champion_trial.resource_lease,
                )
                logger.info(
                    "old champion loss avg %s, loss cb %s",
                    champion_trial.result.loss_avg,
                    champion_trial.result.loss_cb,
                )
                return True
        return False

    @staticmethod
    def _test_avg_loss(champion_trial, trial, warmstart_num=1) -> bool:
        """Comare the challenger(i.e., trial)'s average loss with the
        champion_trial's average loss
        """
        assert trial.trial_id != champion_trial.trial_id
        if trial.result.resource_used >= warmstart_num:
            if trial.result.loss_avg < champion_trial.result.loss_avg:
                logger.info("=====new champion condition satisfied using avg loss=====")
                logger.info("trial %s", trial.config)
                logger.info(
                    "trial loss_avg:%s, trial loss_cb %s",
                    trial.result.loss_avg,
                    trial.result.loss_cb,
                )
                logger.info(
                    "champion loss_avg:%s, champion loss_cb %s",
                    champion_trial.result.loss_avg,
                    champion_trial.result.loss_cb,
                )
                logger.info("champion %s", champion_trial.config)
                return True
        return False

    @staticmethod
    def _test_martingale(champion_trial, trial):
        """Comare the challenger and champion using confidence sequence based
        test martingale

        Not implementated yet
        """
        NotImplementedError



import os

import setuptools

here = os.path.abspath(os.path.dirname(__file__))

with open("README.md", encoding="UTF-8") as fh:
    long_description = fh.read()


# Get the code version
version = {}
with open(os.path.join(here, "flaml/version.py")) as fp:
    exec(fp.read(), version)
__version__ = version["__version__"]

install_requires = [
    "NumPy>=1.17",
]


setuptools.setup(
    name="FLAML",
    version=__version__,
    author="Microsoft Corporation",
    author_email="hpo@microsoft.com",
    description="A fast library for automated machine learning and tuning",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/microsoft/FLAML",
    packages=setuptools.find_packages(include=["flaml*"]),
    package_data={
        "flaml.default": ["*/*.json"],
    },
    include_package_data=True,
    install_requires=install_requires,
    extras_require={
        "automl": [
            "lightgbm>=2.3.1",
            "xgboost>=0.90,<3.0.0",
            "scipy>=1.4.1",
            "pandas>=1.1.4",
            "scikit-learn>=1.0.0",
        ],
        "notebook": [
            "jupyter",
        ],
        "spark": [
            "pyspark>=3.2.0",
            "joblibspark>=0.5.0",
            "joblib<=1.3.2",
        ],
        "test": [
            "jupyter",
            "lightgbm>=2.3.1",
            "xgboost>=0.90,<2.0.0",
            "scipy>=1.4.1",
            "pandas>=1.1.4,<2.0.0; python_version<'3.10'",
            "pandas>=1.1.4; python_version>='3.10'",
            "scikit-learn>=1.0.0",
            "thop",
            "pytest>=6.1.1",
            "coverage>=5.3",
            "pre-commit",
            "torch",
            "torchvision",
            "catboost>=0.26,<1.2; python_version<'3.11'",
            "catboost>=0.26; python_version>='3.11'",
            "rgf-python",
            "optuna>=2.8.0,<=3.6.1",
            "openml",
            "statsmodels>=0.12.2",
            "psutil==5.8.0",
            "dataclasses",
            "transformers[torch]==4.26",
            "datasets<=3.5.0",
            "nltk<=3.8.1",  # 3.8.2 doesn't work with mlflow
            "rouge_score",
            "hcrystalball==0.1.10",
            "seqeval",
            "pytorch-forecasting>=0.9.0,<=0.10.1; python_version<'3.11'",
            # "pytorch-forecasting==0.10.1; python_version=='3.11'",
            "mlflow==2.15.1",
            "joblibspark>=0.5.0",
            "joblib<=1.3.2",
            "nbconvert",
            "nbformat",
            "ipykernel",
            "pytorch-lightning<1.9.1",  # test_forecast_panel
            "tensorboardX==2.6",  # test_forecast_panel
            "requests<2.29.0",  # https://github.com/docker/docker-py/issues/3113
            "packaging",
            "pydantic==1.10.9",
            "sympy",
            "wolframalpha",
            "dill",  # a drop in replacement of pickle
        ],
        "catboost": [
            "catboost>=0.26,<1.2; python_version<'3.11'",
            "catboost>=0.26,<=1.2.5; python_version>='3.11'",
        ],
        "blendsearch": [
            "optuna>=2.8.0,<=3.6.1",
            "packaging",
        ],
        "ray": [
            "ray[tune]~=1.13",
        ],
        "azureml": [
            "azureml-mlflow",
        ],
        "nni": [
            "nni",
        ],
        "vw": [
            "vowpalwabbit>=8.10.0, <9.0.0",
            "scikit-learn",
        ],
        "hf": [
            "transformers[torch]==4.26",
            "datasets",
            "nltk<=3.8.1",
            "rouge_score",
            "seqeval",
        ],
        "nlp": [  # for backward compatibility; hf is the new option name
            "transformers[torch]==4.26",
            "datasets",
            "nltk<=3.8.1",
            "rouge_score",
            "seqeval",
        ],
        "ts_forecast": [
            "holidays<0.14",  # to prevent installation error for prophet
            "prophet>=1.0.1",
            "statsmodels>=0.12.2",
            "hcrystalball==0.1.10",
        ],
        "forecast": [
            "holidays<0.14",  # to prevent installation error for prophet
            "prophet>=1.0.1",
            "statsmodels>=0.12.2",
            "hcrystalball==0.1.10",
            "pytorch-forecasting>=0.9.0; python_version<'3.11'",
            # "pytorch-forecasting==0.10.1; python_version=='3.11'",
            "pytorch-lightning==1.9.0",
            "tensorboardX==2.6",
        ],
        "benchmark": ["catboost>=0.26", "psutil==5.8.0", "xgboost==1.3.3", "pandas==1.1.4"],
        "openai": ["openai==0.27.8", "diskcache"],
        "autogen": ["openai==0.27.8", "diskcache", "termcolor"],
        "mathchat": ["openai==0.27.8", "diskcache", "termcolor", "sympy", "pydantic==1.10.9", "wolframalpha"],
        "retrievechat": [
            "openai==0.27.8",
            "diskcache",
            "termcolor",
            "chromadb",
            "tiktoken",
            "sentence_transformers",
        ],
        "synapse": [
            "joblibspark>=0.5.0",
            "optuna>=2.8.0,<=3.6.1",
            "pyspark>=3.2.0",
        ],
        "autozero": ["scikit-learn", "pandas", "packaging"],
    },
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        # Specify the Python versions you support here.
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    python_requires=">=3.9",
)



# Copyright 2020 The Ray Authors.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This source file is adapted here because ray does not fully support Windows.

# Copyright (c) Microsoft Corporation.
import copy
import functools
import logging
import pickle
import time
import warnings
from collections import defaultdict
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np

from ..sample import (
    Categorical,
    Domain,
    Float,
    Integer,
    LogUniform,
    Quantized,
    Uniform,
)
from ..trial import flatten_dict, unflatten_dict
from .variant_generator import parse_spec_vars

logger = logging.getLogger(__name__)

UNRESOLVED_SEARCH_SPACE = str(
    "You passed a `{par}` parameter to {cls} that contained unresolved search "
    "space definitions. {cls} should however be instantiated with fully "
    "configured search spaces only. To use Ray Tune's automatic search space "
    "conversion, pass the space definition as part of the `config` argument "
    "to `tune.run()` instead."
)

UNDEFINED_SEARCH_SPACE = str(
    "Trying to sample a configuration from {cls}, but no search "
    "space has been defined. Either pass the `{space}` argument when "
    "instantiating the search algorithm, or pass a `config` to "
    "`tune.run()`."
)

UNDEFINED_METRIC_MODE = str(
    "Trying to sample a configuration from {cls}, but the `metric` "
    "({metric}) or `mode` ({mode}) parameters have not been set. "
    "Either pass these arguments when instantiating the search algorithm, "
    "or pass them to `tune.run()`."
)


class Searcher:
    """Abstract class for wrapping suggesting algorithms.
    Custom algorithms can extend this class easily by overriding the
    `suggest` method provide generated parameters for the trials.
    Any subclass that implements ``__init__`` must also call the
    constructor of this class: ``super(Subclass, self).__init__(...)``.
    To track suggestions and their corresponding evaluations, the method
    `suggest` will be passed a trial_id, which will be used in
    subsequent notifications.
    Not all implementations support multi objectives.
    Args:
        metric (str or list): The training result objective value attribute. If
            list then list of training result objective value attributes
        mode (str or list): If string One of {min, max}. If list then
            list of max and min, determines whether objective is minimizing
            or maximizing the metric attribute. Must match type of metric.

    ```python
    class ExampleSearch(Searcher):
        def __init__(self, metric="mean_loss", mode="min", **kwargs):
            super(ExampleSearch, self).__init__(
                metric=metric, mode=mode, **kwargs)
            self.optimizer = Optimizer()
            self.configurations = {}
        def suggest(self, trial_id):
            configuration = self.optimizer.query()
            self.configurations[trial_id] = configuration
        def on_trial_complete(self, trial_id, result, **kwargs):
            configuration = self.configurations[trial_id]
            if result and self.metric in result:
                self.optimizer.update(configuration, result[self.metric])
    tune.run(trainable_function, search_alg=ExampleSearch())
    ```

    """

    FINISHED = "FINISHED"
    CKPT_FILE_TMPL = "searcher-state-{}.pkl"

    def __init__(
        self,
        metric: Optional[str] = None,
        mode: Optional[str] = None,
        max_concurrent: Optional[int] = None,
        use_early_stopped_trials: Optional[bool] = None,
    ):
        self._metric = metric
        self._mode = mode

        if not mode or not metric:
            # Early return to avoid assertions
            return

        assert isinstance(metric, type(mode)), "metric and mode must be of the same type"
        if isinstance(mode, str):
            assert mode in ["min", "max"], "if `mode` is a str must be 'min' or 'max'!"
        elif isinstance(mode, list):
            assert len(mode) == len(metric), "Metric and mode must be the same length"
            assert all(mod in ["min", "max", "obs"] for mod in mode), "All of mode must be 'min' or 'max' or 'obs'!"
        else:
            raise ValueError("Mode must either be a list or string")

    def set_search_properties(self, metric: Optional[str], mode: Optional[str], config: Dict) -> bool:
        """Pass search properties to searcher.
        This method acts as an alternative to instantiating search algorithms
        with their own specific search spaces. Instead they can accept a
        Tune config through this method. A searcher should return ``True``
        if setting the config was successful, or ``False`` if it was
        unsuccessful, e.g. when the search space has already been set.
        Args:
            metric (str): Metric to optimize
            mode (str): One of ["min", "max"]. Direction to optimize.
            config (dict): Tune config dict.
        """
        return False

    def on_trial_result(self, trial_id: str, result: Dict):
        """Optional notification for result during training.
        Note that by default, the result dict may include NaNs or
        may not include the optimization metric. It is up to the
        subclass implementation to preprocess the result to
        avoid breaking the optimization process.
        Args:
            trial_id (str): A unique string ID for the trial.
            result (dict): Dictionary of metrics for current training progress.
                Note that the result dict may include NaNs or
                may not include the optimization metric. It is up to the
                subclass implementation to preprocess the result to
                avoid breaking the optimization process.
        """
        pass

    @property
    def metric(self) -> str:
        """The training result objective value attribute."""
        return self._metric

    @property
    def mode(self) -> str:
        """Specifies if minimizing or maximizing the metric."""
        return self._mode


class ConcurrencyLimiter(Searcher):
    """A wrapper algorithm for limiting the number of concurrent trials.
    Args:
        searcher (Searcher): Searcher object that the
            ConcurrencyLimiter will manage.
        max_concurrent (int): Maximum concurrent samples from the underlying
            searcher.
        batch (bool): Whether to wait for all concurrent samples
            to finish before updating the underlying searcher.
    Example:
    ```python
    from ray.tune.suggest import ConcurrencyLimiter  # ray version < 2
    search_alg = HyperOptSearch(metric="accuracy")
    search_alg = ConcurrencyLimiter(search_alg, max_concurrent=2)
    tune.run(trainable, search_alg=search_alg)
    ```
    """

    def __init__(self, searcher: Searcher, max_concurrent: int, batch: bool = False):
        assert isinstance(max_concurrent, int) and max_concurrent > 0
        self.searcher = searcher
        self.max_concurrent = max_concurrent
        self.batch = batch
        self.live_trials = set()
        self.cached_results = {}
        super().__init__(metric=self.searcher.metric, mode=self.searcher.mode)

    def suggest(self, trial_id: str) -> Optional[Dict]:
        assert trial_id not in self.live_trials, f"Trial ID {trial_id} must be unique: already found in set."
        if len(self.live_trials) >= self.max_concurrent:
            logger.debug(
                f"Not providing a suggestion for {trial_id} due to " "concurrency limit: %s/%s.",
                len(self.live_trials),
                self.max_concurrent,
            )
            return

        suggestion = self.searcher.suggest(trial_id)
        if suggestion not in (None, Searcher.FINISHED):
            self.live_trials.add(trial_id)
        return suggestion

    def on_trial_complete(self, trial_id: str, result: Optional[Dict] = None, error: bool = False):
        if trial_id not in self.live_trials:
            return
        elif self.batch:
            self.cached_results[trial_id] = (result, error)
            if len(self.cached_results) == self.max_concurrent:
                # Update the underlying searcher once the
                # full batch is completed.
                for trial_id, (result, error) in self.cached_results.items():
                    self.searcher.on_trial_complete(trial_id, result=result, error=error)
                    self.live_trials.remove(trial_id)
                self.cached_results = {}
            else:
                return
        else:
            self.searcher.on_trial_complete(trial_id, result=result, error=error)
            self.live_trials.remove(trial_id)

    def get_state(self) -> Dict:
        state = self.__dict__.copy()
        del state["searcher"]
        return copy.deepcopy(state)

    def set_state(self, state: Dict):
        self.__dict__.update(state)

    def save(self, checkpoint_path: str):
        self.searcher.save(checkpoint_path)

    def restore(self, checkpoint_path: str):
        self.searcher.restore(checkpoint_path)

    def on_pause(self, trial_id: str):
        self.searcher.on_pause(trial_id)

    def on_unpause(self, trial_id: str):
        self.searcher.on_unpause(trial_id)

    def set_search_properties(self, metric: Optional[str], mode: Optional[str], config: Dict) -> bool:
        return self.searcher.set_search_properties(metric, mode, config)


try:
    import optuna as ot
    from optuna.distributions import BaseDistribution as OptunaDistribution
    from optuna.samplers import BaseSampler
    from optuna.trial import Trial as OptunaTrial
    from optuna.trial import TrialState as OptunaTrialState
except ImportError:
    ot = None
    OptunaDistribution = None
    BaseSampler = None
    OptunaTrialState = None
    OptunaTrial = None

DEFAULT_METRIC = "_metric"

TRAINING_ITERATION = "training_iteration"

DEFINE_BY_RUN_WARN_THRESHOLD_S = 1


def validate_warmstart(
    parameter_names: List[str],
    points_to_evaluate: List[Union[List, Dict]],
    evaluated_rewards: List,
    validate_point_name_lengths: bool = True,
):
    """Generic validation of a Searcher's warm start functionality.
    Raises exceptions in case of type and length mismatches between
    parameters.
    If ``validate_point_name_lengths`` is False, the equality of lengths
    between ``points_to_evaluate`` and ``parameter_names`` will not be
    validated.
    """
    if points_to_evaluate:
        if not isinstance(points_to_evaluate, list):
            raise TypeError(f"points_to_evaluate expected to be a list, got {type(points_to_evaluate)}.")
        for point in points_to_evaluate:
            if not isinstance(point, (dict, list)):
                raise TypeError(f"points_to_evaluate expected to include list or dict, " f"got {point}.")

            if validate_point_name_lengths and (not len(point) == len(parameter_names)):
                raise ValueError(f"Dim of point {point}" + f" and parameter_names {parameter_names}" + " do not match.")

    if points_to_evaluate and evaluated_rewards:
        if not isinstance(evaluated_rewards, list):
            raise TypeError(f"evaluated_rewards expected to be a list, got {type(evaluated_rewards)}.")
        if not len(evaluated_rewards) == len(points_to_evaluate):
            raise ValueError(
                f"Dim of evaluated_rewards {evaluated_rewards}"
                + f" and points_to_evaluate {points_to_evaluate}"
                + " do not match."
            )


class _OptunaTrialSuggestCaptor:
    """Utility to capture returned values from Optuna's suggest_ methods.

    This will wrap around the ``optuna.Trial` object and decorate all
    `suggest_` callables with a function capturing the returned value,
    which will be saved in the ``captured_values`` dict.
    """

    def __init__(self, ot_trial: OptunaTrial) -> None:
        self.ot_trial = ot_trial
        self.captured_values: Dict[str, Any] = {}

    def _get_wrapper(self, func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # name is always the first arg for suggest_ methods
            name = kwargs.get("name", args[0])
            ret = func(*args, **kwargs)
            self.captured_values[name] = ret
            return ret

        return wrapper

    def __getattr__(self, item_name: str) -> Any:
        item = getattr(self.ot_trial, item_name)
        if item_name.startswith("suggest_") and callable(item):
            return self._get_wrapper(item)
        return item


class OptunaSearch(Searcher):
    """A wrapper around Optuna to provide trial suggestions.

    `Optuna <https://optuna.org/>`_ is a hyperparameter optimization library.
    In contrast to other libraries, it employs define-by-run style
    hyperparameter definitions.

    This Searcher is a thin wrapper around Optuna's search algorithms.
    You can pass any Optuna sampler, which will be used to generate
    hyperparameter suggestions.

    Multi-objective optimization is supported.

    Args:
        space: Hyperparameter search space definition for
            Optuna's sampler. This can be either a dict with
            parameter names as keys and ``optuna.distributions`` as values,
            or a Callable - in which case, it should be a define-by-run
            function using ``optuna.trial`` to obtain the hyperparameter
            values. The function should return either a dict of
            constant values with names as keys, or None.
            For more information, see https://optuna.readthedocs.io\
/en/stable/tutorial/10_key_features/002_configurations.html.

            Warning - No actual computation should take place in the define-by-run
            function. Instead, put the training logic inside the function
            or class trainable passed to ``tune.run``.

        metric: The training result objective value attribute. If
            None but a mode was passed, the anonymous metric ``_metric``
            will be used per default. Can be a list of metrics for
            multi-objective optimization.
        mode: One of {min, max}. Determines whether objective is
            minimizing or maximizing the metric attribute. Can be a list of
            modes for multi-objective optimization (corresponding to
            ``metric``).
        points_to_evaluate: Initial parameter suggestions to be run
            first. This is for when you already have some good parameters
            you want to run first to help the algorithm make better suggestions
            for future parameters. Needs to be a list of dicts containing the
            configurations.
        sampler: Optuna sampler used to
            draw hyperparameter configurations. Defaults to ``MOTPESampler``
            for multi-objective optimization with Optuna<2.9.0, and
            ``TPESampler`` in every other case.

            Warning: Please note that with Optuna 2.10.0 and earlier
                default ``MOTPESampler``/``TPESampler`` suffer
                from performance issues when dealing with a large number of
                completed trials (approx. >100). This will manifest as
                a delay when suggesting new configurations.
                This is an Optuna issue and may be fixed in a future
                Optuna release.

        seed: Seed to initialize sampler with. This parameter is only
            used when ``sampler=None``. In all other cases, the sampler
            you pass should be initialized with the seed already.
        evaluated_rewards: If you have previously evaluated the
            parameters passed in as points_to_evaluate you can avoid
            re-running those trials by passing in the reward attributes
            as a list so the optimiser can be told the results without
            needing to re-compute the trial. Must be the same length as
            points_to_evaluate.

            Warning - When using ``evaluated_rewards``, the search space ``space``
            must be provided as a dict with parameter names as
            keys and ``optuna.distributions`` instances as values. The
            define-by-run search space definition is not yet supported with
            this functionality.

    Tune automatically converts search spaces to Optuna's format:

    ```python
    from ray.tune.suggest.optuna import OptunaSearch

    config = {
        "a": tune.uniform(6, 8)
        "b": tune.loguniform(1e-4, 1e-2)
    }

    optuna_search = OptunaSearch(
        metric="loss",
        mode="min")

    tune.run(trainable, config=config, search_alg=optuna_search)
    ```

    If you would like to pass the search space manually, the code would
    look like this:

    ```python
    from ray.tune.suggest.optuna import OptunaSearch
    import optuna

    space = {
        "a": optuna.distributions.UniformDistribution(6, 8),
        "b": optuna.distributions.LogUniformDistribution(1e-4, 1e-2),
    }

    optuna_search = OptunaSearch(
        space,
        metric="loss",
        mode="min")

    tune.run(trainable, search_alg=optuna_search)

    # Equivalent Optuna define-by-run function approach:

    def define_search_space(trial: optuna.Trial):
        trial.suggest_float("a", 6, 8)
        trial.suggest_float("b", 1e-4, 1e-2, log=True)
        # training logic goes into trainable, this is just
        # for search space definition

    optuna_search = OptunaSearch(
        define_search_space,
        metric="loss",
        mode="min")

    tune.run(trainable, search_alg=optuna_search)
    ```

    Multi-objective optimization is supported:

    ```python
    from ray.tune.suggest.optuna import OptunaSearch
    import optuna

    space = {
        "a": optuna.distributions.UniformDistribution(6, 8),
        "b": optuna.distributions.LogUniformDistribution(1e-4, 1e-2),
    }

    # Note you have to specify metric and mode here instead of
    # in tune.run
    optuna_search = OptunaSearch(
        space,
        metric=["loss1", "loss2"],
        mode=["min", "max"])

    # Do not specify metric and mode here!
    tune.run(
        trainable,
        search_alg=optuna_search
    )
    ```

    You can pass configs that will be evaluated first using
    ``points_to_evaluate``:

    ```python
    from ray.tune.suggest.optuna import OptunaSearch
    import optuna

    space = {
        "a": optuna.distributions.UniformDistribution(6, 8),
        "b": optuna.distributions.LogUniformDistribution(1e-4, 1e-2),
    }

    optuna_search = OptunaSearch(
        space,
        points_to_evaluate=[{"a": 6.5, "b": 5e-4}, {"a": 7.5, "b": 1e-3}]
        metric="loss",
        mode="min")

    tune.run(trainable, search_alg=optuna_search)
    ```

    Avoid re-running evaluated trials by passing the rewards together with
    `points_to_evaluate`:

    ```python
    from ray.tune.suggest.optuna import OptunaSearch
    import optuna

    space = {
        "a": optuna.distributions.UniformDistribution(6, 8),
        "b": optuna.distributions.LogUniformDistribution(1e-4, 1e-2),
    }

    optuna_search = OptunaSearch(
        space,
        points_to_evaluate=[{"a": 6.5, "b": 5e-4}, {"a": 7.5, "b": 1e-3}]
        evaluated_rewards=[0.89, 0.42]
        metric="loss",
        mode="min")

    tune.run(trainable, search_alg=optuna_search)
    ```

    """

    def __init__(
        self,
        space: Optional[
            Union[
                Dict[str, "OptunaDistribution"],
                List[Tuple],
                Callable[["OptunaTrial"], Optional[Dict[str, Any]]],
            ]
        ] = None,
        metric: Optional[Union[str, List[str]]] = None,
        mode: Optional[Union[str, List[str]]] = None,
        points_to_evaluate: Optional[List[Dict]] = None,
        sampler: Optional["BaseSampler"] = None,
        seed: Optional[int] = None,
        evaluated_rewards: Optional[List] = None,
    ):
        assert ot is not None, "Optuna must be installed! Run `pip install optuna`."
        super().__init__(metric=metric, mode=mode)

        if isinstance(space, dict) and space:
            resolved_vars, domain_vars, grid_vars = parse_spec_vars(space)
            if domain_vars or grid_vars:
                logger.warning(UNRESOLVED_SEARCH_SPACE.format(par="space", cls=type(self).__name__))
                space = self.convert_search_space(space)
            else:
                # Flatten to support nested dicts
                space = flatten_dict(space, "/")

        self._space = space

        self._points_to_evaluate = points_to_evaluate or []
        # rewards should be a list of floats, not a dict
        # After Optuna > 3.5.0, there is a check for NaN in the list "any(math.isnan(x) for x in self._values)"
        # which will raise an error when encountering a dict
        if evaluated_rewards is not None:
            self._evaluated_rewards = [
                list(item.values())[0] if isinstance(item, dict) else item for item in evaluated_rewards
            ]
        else:
            self._evaluated_rewards = evaluated_rewards

        self._study_name = "optuna"  # Fixed study name for in-memory storage

        if sampler and seed:
            logger.warning(
                "You passed an initialized sampler to `OptunaSearch`. The "
                "`seed` parameter has to be passed to the sampler directly "
                "and will be ignored."
            )
        elif sampler:
            assert isinstance(sampler, BaseSampler), (
                "You can only pass an instance of " "`optuna.samplers.BaseSampler` " "as a sampler to `OptunaSearcher`."
            )

        self._sampler = sampler
        self._seed = seed

        self._completed_trials = set()

        self._ot_trials = {}
        self._ot_study = None
        if self._space:
            self._setup_study(mode)

    def _setup_study(self, mode: Union[str, list]):
        if self._metric is None and self._mode:
            if isinstance(self._mode, list):
                raise ValueError(
                    "If ``mode`` is a list (multi-objective optimization " "case), ``metric`` must be defined."
                )
            # If only a mode was passed, use anonymous metric
            self._metric = DEFAULT_METRIC

        pruner = ot.pruners.NopPruner()
        storage = ot.storages.InMemoryStorage()
        try:
            from packaging import version
        except ImportError:
            raise ImportError("To use BlendSearch, run: pip install flaml[blendsearch]")
        if self._sampler:
            sampler = self._sampler
        elif isinstance(mode, list) and version.parse(ot.__version__) < version.parse("2.9.0"):
            # MOTPESampler deprecated in Optuna>=2.9.0
            sampler = ot.samplers.MOTPESampler(seed=self._seed)
        else:
            sampler = ot.samplers.TPESampler(seed=self._seed)

        if isinstance(mode, list):
            study_direction_args = dict(
                directions=["minimize" if m == "min" else "maximize" for m in mode],
            )
        else:
            study_direction_args = dict(
                direction="minimize" if mode == "min" else "maximize",
            )

        self._ot_study = ot.study.create_study(
            storage=storage,
            sampler=sampler,
            pruner=pruner,
            study_name=self._study_name,
            load_if_exists=True,
            **study_direction_args,
        )

        if self._points_to_evaluate:
            validate_warmstart(
                self._space,
                self._points_to_evaluate,
                self._evaluated_rewards,
                validate_point_name_lengths=not callable(self._space),
            )
            if self._evaluated_rewards:
                for point, reward in zip(self._points_to_evaluate, self._evaluated_rewards):
                    self.add_evaluated_point(point, reward)
            else:
                for point in self._points_to_evaluate:
                    self._ot_study.enqueue_trial(point)

    def set_search_properties(self, metric: Optional[str], mode: Optional[str], config: Dict, **spec) -> bool:
        if self._space:
            return False
        space = self.convert_search_space(config)
        self._space = space
        if metric:
            self._metric = metric
        if mode:
            self._mode = mode

        self._setup_study(self._mode)
        return True

    def _suggest_from_define_by_run_func(
        self,
        func: Callable[["OptunaTrial"], Optional[Dict[str, Any]]],
        ot_trial: "OptunaTrial",
    ) -> Dict:
        captor = _OptunaTrialSuggestCaptor(ot_trial)
        time_start = time.time()
        ret = func(captor)
        time_taken = time.time() - time_start
        if time_taken > DEFINE_BY_RUN_WARN_THRESHOLD_S:
            warnings.warn(
                "Define-by-run function passed in the `space` argument "
                f"took {time_taken} seconds to "
                "run. Ensure that actual computation, training takes "
                "place inside Tune's train functions or Trainables "
                "passed to `tune.run`."
            )
        if ret is not None:
            if not isinstance(ret, dict):
                raise TypeError(
                    "The return value of the define-by-run function "
                    "passed in the `space` argument should be "
                    "either None or a `dict` with `str` keys. "
                    f"Got {type(ret)}."
                )
            if not all(isinstance(k, str) for k in ret.keys()):
                raise TypeError(
                    "At least one of the keys in the dict returned by the "
                    "define-by-run function passed in the `space` argument "
                    "was not a `str`."
                )
        return {**captor.captured_values, **ret} if ret else captor.captured_values

    def suggest(self, trial_id: str) -> Optional[Dict]:
        if not self._space:
            raise RuntimeError(UNDEFINED_SEARCH_SPACE.format(cls=self.__class__.__name__, space="space"))
        if not self._metric or not self._mode:
            raise RuntimeError(
                UNDEFINED_METRIC_MODE.format(cls=self.__class__.__name__, metric=self._metric, mode=self._mode)
            )
        if callable(self._space):
            # Define-by-run case
            if trial_id not in self._ot_trials:
                self._ot_trials[trial_id] = self._ot_study.ask()

            ot_trial = self._ot_trials[trial_id]

            params = self._suggest_from_define_by_run_func(self._space, ot_trial)
        else:
            # Use Optuna ask interface (since version 2.6.0)
            if trial_id not in self._ot_trials:
                self._ot_trials[trial_id] = self._ot_study.ask(fixed_distributions=self._space)
            ot_trial = self._ot_trials[trial_id]
            params = ot_trial.params

        return unflatten_dict(params)

    def on_trial_result(self, trial_id: str, result: Dict):
        if isinstance(self.metric, list):
            # Optuna doesn't support incremental results
            # for multi-objective optimization
            return
        if trial_id in self._completed_trials:
            logger.warning(
                f"Received additional result for trial {trial_id}, but " f"it already finished. Result: {result}"
            )
            return
        metric = result[self.metric]
        step = result[TRAINING_ITERATION]
        ot_trial = self._ot_trials[trial_id]
        ot_trial.report(metric, step)

    def on_trial_complete(self, trial_id: str, result: Optional[Dict] = None, error: bool = False):
        if trial_id in self._completed_trials:
            logger.warning(
                f"Received additional completion for trial {trial_id}, but " f"it already finished. Result: {result}"
            )
            return

        ot_trial = self._ot_trials[trial_id]

        if result:
            if isinstance(self.metric, list):
                val = [result.get(metric, None) for metric in self.metric]
            else:
                val = result.get(self.metric, None)
        else:
            val = None
        ot_trial_state = OptunaTrialState.COMPLETE
        if val is None:
            if error:
                ot_trial_state = OptunaTrialState.FAIL
            else:
                ot_trial_state = OptunaTrialState.PRUNED
        try:
            self._ot_study.tell(ot_trial, val, state=ot_trial_state)
        except Exception as exc:
            logger.warning(exc)  # E.g. if NaN was reported

        self._completed_trials.add(trial_id)

    def add_evaluated_point(
        self,
        parameters: Dict,
        value: float,
        error: bool = False,
        pruned: bool = False,
        intermediate_values: Optional[List[float]] = None,
    ):
        if not self._space:
            raise RuntimeError(UNDEFINED_SEARCH_SPACE.format(cls=self.__class__.__name__, space="space"))
        if not self._metric or not self._mode:
            raise RuntimeError(
                UNDEFINED_METRIC_MODE.format(cls=self.__class__.__name__, metric=self._metric, mode=self._mode)
            )
        if callable(self._space):
            raise TypeError(
                "Define-by-run function passed in `space` argument is not "
                "yet supported when using `evaluated_rewards`. Please provide "
                "an `OptunaDistribution` dict or pass a Ray Tune "
                "search space to `tune.run()`."
            )

        ot_trial_state = OptunaTrialState.COMPLETE
        if error:
            ot_trial_state = OptunaTrialState.FAIL
        elif pruned:
            ot_trial_state = OptunaTrialState.PRUNED

        if intermediate_values:
            intermediate_values_dict = {i: value for i, value in enumerate(intermediate_values)}
        else:
            intermediate_values_dict = None

        trial = ot.trial.create_trial(
            state=ot_trial_state,
            value=value,
            params=parameters,
            distributions=self._space,
            intermediate_values=intermediate_values_dict,
        )

        self._ot_study.add_trial(trial)

    def save(self, checkpoint_path: str):
        save_object = (
            self._sampler,
            self._ot_trials,
            self._ot_study,
            self._points_to_evaluate,
            self._evaluated_rewards,
        )
        with open(checkpoint_path, "wb") as outputFile:
            pickle.dump(save_object, outputFile)

    def restore(self, checkpoint_path: str):
        with open(checkpoint_path, "rb") as inputFile:
            save_object = pickle.load(inputFile)
        if len(save_object) == 5:
            (
                self._sampler,
                self._ot_trials,
                self._ot_study,
                self._points_to_evaluate,
                self._evaluated_rewards,
            ) = save_object
        else:
            # Backwards compatibility
            (
                self._sampler,
                self._ot_trials,
                self._ot_study,
                self._points_to_evaluate,
            ) = save_object

    @staticmethod
    def convert_search_space(spec: Dict) -> Dict[str, Any]:
        resolved_vars, domain_vars, grid_vars = parse_spec_vars(spec)

        if not domain_vars and not grid_vars:
            return {}

        if grid_vars:
            raise ValueError("Grid search parameters cannot be automatically converted " "to an Optuna search space.")

        # Flatten and resolve again after checking for grid search.
        spec = flatten_dict(spec, prevent_delimiter=True)
        resolved_vars, domain_vars, grid_vars = parse_spec_vars(spec)

        def resolve_value(domain: Domain) -> ot.distributions.BaseDistribution:
            quantize = None

            sampler = domain.get_sampler()
            if isinstance(sampler, Quantized):
                quantize = sampler.q
                sampler = sampler.sampler
                if isinstance(sampler, LogUniform):
                    logger.warning(
                        "Optuna does not handle quantization in loguniform "
                        "sampling. The parameter will be passed but it will "
                        "probably be ignored."
                    )

            if isinstance(domain, Float):
                if isinstance(sampler, LogUniform):
                    if quantize:
                        logger.warning(
                            "Optuna does not support both quantization and "
                            "sampling from LogUniform. Dropped quantization."
                        )
                    return ot.distributions.LogUniformDistribution(domain.lower, domain.upper)

                elif isinstance(sampler, Uniform):
                    if quantize:
                        return ot.distributions.DiscreteUniformDistribution(domain.lower, domain.upper, quantize)
                    return ot.distributions.UniformDistribution(domain.lower, domain.upper)

            elif isinstance(domain, Integer):
                if isinstance(sampler, LogUniform):
                    # ``step`` argument Deprecated in v2.0.0. ``step`` argument should be 1 in Log Distribution
                    # The removal of this feature is currently scheduled for v4.0.0,
                    return ot.distributions.IntLogUniformDistribution(domain.lower, domain.upper - 1, step=1)
                elif isinstance(sampler, Uniform):
                    # Upper bound should be inclusive for quantization and
                    # exclusive otherwise
                    return ot.distributions.IntUniformDistribution(
                        domain.lower,
                        domain.upper - int(bool(not quantize)),
                        step=quantize or 1,
                    )
            elif isinstance(domain, Categorical):
                if isinstance(sampler, Uniform):
                    return ot.distributions.CategoricalDistribution(domain.categories)

            raise ValueError(
                "Optuna search does not support parameters of type "
                "`{}` with samplers of type `{}`".format(type(domain).__name__, type(domain.sampler).__name__)
            )

        # Parameter name is e.g. "a/b/c" for nested dicts
        values = {"/".join(path): resolve_value(domain) for path, domain in domain_vars}

        return values



# This file is adapted from
# https://github.com/locuslab/TCN/blob/master/TCN/tcn.py
# https://github.com/locuslab/TCN/blob/master/TCN/adding_problem/add_test.py

import datetime
import logging
import time

import pandas as pd
import pytorch_lightning as pl
import torch
import torch.nn as nn
import torch.optim as optim
from pytorch_lightning.callbacks import EarlyStopping, LearningRateMonitor
from pytorch_lightning.loggers import TensorBoardLogger
from torch.nn.utils import weight_norm
from torch.utils.data import DataLoader, TensorDataset

from flaml import tune
from flaml.automl.data import add_time_idx_col
from flaml.automl.logger import logger, logger_formatter
from flaml.automl.time_series.ts_data import TimeSeriesDataset
from flaml.automl.time_series.ts_model import TimeSeriesEstimator


class Chomp1d(nn.Module):
    def __init__(self, chomp_size):
        super().__init__()
        self.chomp_size = chomp_size

    def forward(self, x):
        return x[:, :, : -self.chomp_size].contiguous()


class TemporalBlock(nn.Module):
    def __init__(self, n_inputs, n_outputs, kernel_size, stride, dilation, padding, dropout=0.2):
        super().__init__()
        self.conv1 = weight_norm(
            nn.Conv1d(n_inputs, n_outputs, kernel_size, stride=stride, padding=padding, dilation=dilation)
        )
        self.chomp1 = Chomp1d(padding)
        self.relu1 = nn.ReLU()
        self.dropout1 = nn.Dropout(dropout)

        self.conv2 = weight_norm(
            nn.Conv1d(n_outputs, n_outputs, kernel_size, stride=stride, padding=padding, dilation=dilation)
        )
        self.chomp2 = Chomp1d(padding)
        self.relu2 = nn.ReLU()
        self.dropout2 = nn.Dropout(dropout)

        self.net = nn.Sequential(
            self.conv1, self.chomp1, self.relu1, self.dropout1, self.conv2, self.chomp2, self.relu2, self.dropout2
        )
        self.downsample = nn.Conv1d(n_inputs, n_outputs, 1) if n_inputs != n_outputs else None
        self.relu = nn.ReLU()
        self.init_weights()

    def init_weights(self):
        self.conv1.weight.data.normal_(0, 0.01)
        self.conv2.weight.data.normal_(0, 0.01)
        if self.downsample is not None:
            self.downsample.weight.data.normal_(0, 0.01)

    def forward(self, x):
        out = self.net(x)
        res = x if self.downsample is None else self.downsample(x)
        return self.relu(out + res)


class TCNForecaster(nn.Module):
    def __init__(
        self,
        input_feature_num,
        num_outputs,
        num_channels,
        kernel_size=2,
        dropout=0.2,
    ):
        super().__init__()
        layers = []
        num_levels = len(num_channels)
        for i in range(num_levels):
            dilation_size = 2**i
            in_channels = input_feature_num if i == 0 else num_channels[i - 1]
            out_channels = num_channels[i]
            layers += [
                TemporalBlock(
                    in_channels,
                    out_channels,
                    kernel_size,
                    stride=1,
                    dilation=dilation_size,
                    padding=(kernel_size - 1) * dilation_size,
                    dropout=dropout,
                )
            ]

        self.network = nn.Sequential(*layers)
        self.linear = nn.Linear(num_channels[-1], num_outputs)

    def forward(self, x):
        y1 = self.network(x)
        return self.linear(y1[:, :, -1])


class TCNForecasterLightningModule(pl.LightningModule):
    def __init__(self, model: TCNForecaster, learning_rate: float = 1e-3):
        super().__init__()
        self.model = model
        self.learning_rate = learning_rate
        self.loss_fn = nn.MSELoss()

    def forward(self, x):
        return self.model(x)

    def step(self, batch, batch_idx):
        x, y = batch
        y_hat = self.model(x)
        loss = self.loss_fn(y_hat, y)
        return loss

    def training_step(self, batch, batch_idx):
        loss = self.step(batch, batch_idx)
        self.log("train_loss", loss)
        return loss

    def validation_step(self, batch, batch_idx):
        loss = self.step(batch, batch_idx)
        self.log("val_loss", loss)
        return loss

    def configure_optimizers(self):
        return torch.optim.Adam(self.parameters(), lr=self.learning_rate)


class DataframeDataset(torch.utils.data.Dataset):
    def __init__(self, dataframe, target_column, features_columns, sequence_length, train=True):
        self.data = torch.tensor(dataframe[features_columns].to_numpy(), dtype=torch.float)
        self.sequence_length = sequence_length
        if train:
            self.labels = torch.tensor(dataframe[target_column].to_numpy(), dtype=torch.float)
        self.is_train = train

    def __len__(self):
        return len(self.data) - self.sequence_length + 1

    def __getitem__(self, idx):
        data = self.data[idx : idx + self.sequence_length]
        data = data.permute(1, 0)
        if self.is_train:
            label = self.labels[idx : idx + self.sequence_length]
            return data, label
        else:
            return data


class TCNEstimator(TimeSeriesEstimator):
    """The class for tuning TCN Forecaster"""

    @classmethod
    def search_space(cls, data, task, pred_horizon, **params):
        space = {
            "num_levels": {
                "domain": tune.randint(lower=4, upper=20),  # hidden = 2^num_hidden
                "init_value": 4,
            },
            "num_hidden": {
                "domain": tune.randint(lower=4, upper=8),  # hidden = 2^num_hidden
                "init_value": 5,
            },
            "kernel_size": {
                "domain": tune.choice([2, 3, 5, 7]),  # common choices for kernel size
                "init_value": 3,
            },
            "dropout": {
                "domain": tune.uniform(lower=0.0, upper=0.5),  # standard range for dropout
                "init_value": 0.1,
            },
            "learning_rate": {
                "domain": tune.loguniform(lower=1e-4, upper=1e-1),  # typical range for learning rate
                "init_value": 1e-3,
            },
        }
        return space

    def __init__(self, task="ts_forecast", n_jobs=1, **params):
        super().__init__(task, **params)
        logging.getLogger("pytorch_lightning").setLevel(logging.WARNING)

    def fit(self, X_train: TimeSeriesDataset, y_train=None, budget=None, **kwargs):
        start_time = time.time()
        if budget is not None:
            deltabudget = datetime.timedelta(seconds=budget)
        else:
            deltabudget = None
        X_train = self.enrich(X_train)
        super().fit(X_train, y_train, budget, **kwargs)

        self.batch_size = kwargs.get("batch_size", 64)
        self.horizon = kwargs.get("period", 1)
        self.feature_cols = X_train.time_varying_known_reals
        self.target_col = X_train.target_names[0]

        train_dataset = DataframeDataset(
            X_train.train_data,
            self.target_col,
            self.feature_cols,
            self.horizon,
        )
        train_loader = DataLoader(train_dataset, batch_size=self.batch_size, shuffle=False)
        if not X_train.test_data.empty:
            val_dataset = DataframeDataset(
                X_train.test_data,
                self.target_col,
                self.feature_cols,
                self.horizon,
            )
        else:
            val_dataset = DataframeDataset(
                X_train.train_data.sample(frac=0.2, random_state=kwargs.get("random_state", 0)),
                self.target_col,
                self.feature_cols,
                self.horizon,
            )

        val_loader = DataLoader(val_dataset, batch_size=self.batch_size, shuffle=False)

        model = TCNForecaster(
            len(self.feature_cols),
            self.horizon,
            [2 ** self.params["num_hidden"]] * self.params["num_levels"],
            self.params["kernel_size"],
            self.params["dropout"],
        )

        pl_module = TCNForecasterLightningModule(model, self.params["learning_rate"])

        # Training loop
        # gpus is deprecated in v1.7 and removed in v2.0
        # accelerator="auto" can cast all condition.
        trainer = pl.Trainer(
            max_epochs=kwargs.get("max_epochs", 10),
            accelerator="auto",
            callbacks=[
                EarlyStopping(monitor="val_loss", min_delta=1e-4, patience=10, verbose=False, mode="min"),
                LearningRateMonitor(),
            ],
            logger=TensorBoardLogger(kwargs.get("log_dir", "logs/lightning_logs")),  # logging results to a tensorboard
            max_time=deltabudget,
            enable_model_summary=False,
            enable_progress_bar=False,
        )
        trainer.fit(
            pl_module,
            train_dataloaders=train_loader,
            val_dataloaders=val_loader,
        )
        best_model = trainer.model
        self._model = best_model
        train_time = time.time() - start_time
        return train_time

    def predict(self, X):
        X = self.enrich(X)
        if isinstance(X, TimeSeriesDataset):
            df = X.X_val
        else:
            df = X
        dataset = DataframeDataset(
            df,
            self.target_col,
            self.feature_cols,
            self.horizon,
            train=False,
        )
        data_loader = DataLoader(dataset, batch_size=self.batch_size, shuffle=False)
        self._model.eval()
        raw_preds = []
        for batch_x in data_loader:
            raw_pred = self._model(batch_x)
            raw_preds.append(raw_pred)
        raw_preds = torch.cat(raw_preds, dim=0)
        preds = pd.Series(raw_preds.detach().numpy().ravel())
        return preds



import json
import logging
import os
from typing import Dict, List, Optional, Set, Union

NON_CACHE_KEY = ["api_key", "api_base", "api_type", "api_version"]


def get_key(config):
    """Get a unique identifier of a configuration.

    Args:
        config (dict or list): A configuration.

    Returns:
        tuple: A unique identifier which can be used as a key for a dict.
    """
    copied = False
    for key in NON_CACHE_KEY:
        if key in config:
            config, copied = config.copy() if not copied else config, True
            config.pop(key)
    # if isinstance(config, dict):
    #     return tuple(get_key(x) for x in sorted(config.items()))
    # if isinstance(config, list):
    #     return tuple(get_key(x) for x in config)
    # return config
    return json.dumps(config, sort_keys=True)


def get_config_list(
    api_keys: List, api_bases: Optional[List] = None, api_type: Optional[str] = None, api_version: Optional[str] = None
) -> List[Dict]:
    """Get a list of configs for openai api calls.

    Args:
        api_keys (list): The api keys for openai api calls.
        api_bases (list, optional): The api bases for openai api calls.
        api_type (str, optional): The api type for openai api calls.
        api_version (str, optional): The api version for openai api calls.
    """
    config_list = []
    for i, api_key in enumerate(api_keys):
        if not api_key.strip():
            continue
        config = {"api_key": api_key}
        if api_bases:
            config["api_base"] = api_bases[i]
        if api_type:
            config["api_type"] = api_type
        if api_version:
            config["api_version"] = api_version
        config_list.append(config)
    return config_list


def config_list_openai_aoai(
    key_file_path: Optional[str] = ".",
    openai_api_key_file: Optional[str] = "key_openai.txt",
    aoai_api_key_file: Optional[str] = "key_aoai.txt",
    aoai_api_base_file: Optional[str] = "base_aoai.txt",
    exclude: Optional[str] = None,
) -> List[Dict]:
    """Get a list of configs for openai + azure openai api calls.

    Args:
        key_file_path (str, optional): The path to the key files.
        openai_api_key_file (str, optional): The file name of the openai api key.
        aoai_api_key_file (str, optional): The file name of the azure openai api key.
        aoai_api_base_file (str, optional): The file name of the azure openai api base.
        exclude (str, optional): The api type to exclude, "openai" or "aoai".

    Returns:
        list: A list of configs for openai api calls.
    """
    if "OPENAI_API_KEY" not in os.environ and exclude != "openai":
        try:
            with open(f"{key_file_path}/{openai_api_key_file}") as key_file:
                os.environ["OPENAI_API_KEY"] = key_file.read().strip()
        except FileNotFoundError:
            logging.info(
                "To use OpenAI API, please set OPENAI_API_KEY in os.environ "
                "or create key_openai.txt in the specified path, or specify the api_key in config_list."
            )
    if "AZURE_OPENAI_API_KEY" not in os.environ and exclude != "aoai":
        try:
            with open(f"{key_file_path}/{aoai_api_key_file}") as key_file:
                os.environ["AZURE_OPENAI_API_KEY"] = key_file.read().strip()
        except FileNotFoundError:
            logging.info(
                "To use Azure OpenAI API, please set AZURE_OPENAI_API_KEY in os.environ "
                "or create key_aoai.txt in the specified path, or specify the api_key in config_list."
            )
    if "AZURE_OPENAI_API_BASE" not in os.environ and exclude != "aoai":
        try:
            with open(f"{key_file_path}/{aoai_api_base_file}") as key_file:
                os.environ["AZURE_OPENAI_API_BASE"] = key_file.read().strip()
        except FileNotFoundError:
            logging.info(
                "To use Azure OpenAI API, please set AZURE_OPENAI_API_BASE in os.environ "
                "or create base_aoai.txt in the specified path, or specify the api_base in config_list."
            )
    aoai_config = (
        get_config_list(
            # Assuming Azure OpenAI api keys in os.environ["AZURE_OPENAI_API_KEY"], in separated lines
            api_keys=os.environ.get("AZURE_OPENAI_API_KEY", "").split("\n"),
            # Assuming Azure OpenAI api bases in os.environ["AZURE_OPENAI_API_BASE"], in separated lines
            api_bases=os.environ.get("AZURE_OPENAI_API_BASE", "").split("\n"),
            api_type="azure",
            api_version="2023-06-01-preview",  # change if necessary
        )
        if exclude != "aoai"
        else []
    )
    openai_config = (
        get_config_list(
            # Assuming OpenAI API_KEY in os.environ["OPENAI_API_KEY"]
            api_keys=os.environ.get("OPENAI_API_KEY", "").split("\n"),
            # "api_type": "open_ai",
            # "api_base": "https://api.openai.com/v1",
        )
        if exclude != "openai"
        else []
    )
    config_list = openai_config + aoai_config
    return config_list


def config_list_from_models(
    key_file_path: Optional[str] = ".",
    openai_api_key_file: Optional[str] = "key_openai.txt",
    aoai_api_key_file: Optional[str] = "key_aoai.txt",
    aoai_api_base_file: Optional[str] = "base_aoai.txt",
    exclude: Optional[str] = None,
    model_list: Optional[list] = None,
) -> List[Dict]:
    """Get a list of configs for api calls with models in the model list.

    Args:
        key_file_path (str, optional): The path to the key files.
        openai_api_key_file (str, optional): The file name of the openai api key.
        aoai_api_key_file (str, optional): The file name of the azure openai api key.
        aoai_api_base_file (str, optional): The file name of the azure openai api base.
        exclude (str, optional): The api type to exclude, "openai" or "aoai".
        model_list (list, optional): The model list.

    Returns:
        list: A list of configs for openai api calls.
    """
    config_list = config_list_openai_aoai(
        key_file_path,
        openai_api_key_file,
        aoai_api_key_file,
        aoai_api_base_file,
        exclude,
    )
    if model_list:
        config_list = [{**config, "model": model} for model in model_list for config in config_list]
    return config_list


def config_list_gpt4_gpt35(
    key_file_path: Optional[str] = ".",
    openai_api_key_file: Optional[str] = "key_openai.txt",
    aoai_api_key_file: Optional[str] = "key_aoai.txt",
    aoai_api_base_file: Optional[str] = "base_aoai.txt",
    exclude: Optional[str] = None,
) -> List[Dict]:
    """Get a list of configs for gpt-4 followed by gpt-3.5 api calls.

    Args:
        key_file_path (str, optional): The path to the key files.
        openai_api_key_file (str, optional): The file name of the openai api key.
        aoai_api_key_file (str, optional): The file name of the azure openai api key.
        aoai_api_base_file (str, optional): The file name of the azure openai api base.
        exclude (str, optional): The api type to exclude, "openai" or "aoai".

    Returns:
        list: A list of configs for openai api calls.
    """
    return config_list_from_models(
        key_file_path,
        openai_api_key_file,
        aoai_api_key_file,
        aoai_api_base_file,
        exclude,
        model_list=["gpt-4", "gpt-3.5-turbo"],
    )


def filter_config(config_list, filter_dict):
    """Filter the config list by provider and model.

    Args:
        config_list (list): The config list.
        filter_dict (dict, optional): The filter dict with keys corresponding to a field in each config,
            and values corresponding to lists of acceptable values for each key.

    Returns:
        list: The filtered config list.
    """
    if filter_dict:
        config_list = [
            config for config in config_list if all(config.get(key) in value for key, value in filter_dict.items())
        ]
    return config_list


def config_list_from_json(
    env_or_file: str,
    file_location: Optional[str] = "",
    filter_dict: Optional[Dict[str, Union[List[Union[str, None]], Set[Union[str, None]]]]] = None,
) -> List[Dict]:
    """Get a list of configs from a json parsed from an env variable or a file.

    Args:
        env_or_file (str): The env variable name or file name.
        file_location (str, optional): The file location.
        filter_dict (dict, optional): The filter dict with keys corresponding to a field in each config,
            and values corresponding to lists of acceptable values for each key.
            e.g.,
    ```python
    filter_dict = {
        "api_type": ["open_ai", None],  # None means a missing key is acceptable
        "model": ["gpt-3.5-turbo", "gpt-4"],
    }
    ```

    Returns:
        list: A list of configs for openai api calls.
    """
    json_str = os.environ.get(env_or_file)
    if json_str:
        config_list = json.loads(json_str)
    else:
        try:
            with open(os.path.join(file_location, env_or_file)) as json_file:
                config_list = json.load(json_file)
        except FileNotFoundError:
            return []
    return filter_config(config_list, filter_dict)



import copy
import datetime
import math
from dataclasses import dataclass, field
from typing import Callable, Dict, Generator, List, Optional, Union

import numpy as np

try:
    import pandas as pd
    from pandas import DataFrame, Series, to_datetime
    from scipy.sparse import issparse
    from sklearn.compose import ColumnTransformer
    from sklearn.impute import SimpleImputer
    from sklearn.preprocessing import LabelEncoder

    from .feature import monthly_fourier_features
except ImportError:

    class PD:
        pass

    pd = PD()
    pd.DataFrame = None
    pd.Series = None
    DataFrame = Series = None


# dataclass will remove empty default value even with field(default_factory=lambda: [])
# Change into default=None to place the attr
@dataclass
class TimeSeriesDataset:
    train_data: pd.DataFrame
    time_idx: str
    time_col: str
    target_names: List[str]
    frequency: str
    test_data: pd.DataFrame
    time_varying_known_categoricals: List[str] = field(default=None)
    time_varying_known_reals: List[str] = field(default=None)
    time_varying_unknown_categoricals: List[str] = field(default=None)
    time_varying_unknown_reals: List[str] = field(default=None)

    def __init__(
        self,
        train_data: pd.DataFrame,
        time_col: str,
        target_names: Union[str, List[str]],
        time_idx: str = "time_idx",
        test_data: Optional[pd.DataFrame] = None,
    ):
        self.train_data = train_data
        self.time_col = time_col
        self.time_idx = time_idx
        self.target_names = [target_names] if isinstance(target_names, str) else list(target_names)
        assert isinstance(self.target_names, list)
        assert len(self.target_names)

        self.frequency = pd.infer_freq(train_data[time_col].unique())
        assert self.frequency is not None, "Only time series of regular frequency are currently supported."

        float_cols = list(train_data.select_dtypes(include=["floating"]).columns)
        self.time_varying_known_reals = list(set(float_cols) - set(self.target_names))

        self.time_varying_known_categoricals = list(
            set(train_data.columns) - set(self.time_varying_known_reals) - set(self.target_names) - {time_col}
        )
        if test_data is not None:
            self.test_data = test_data
        else:
            self.test_data = pd.DataFrame(columns=self.train_data.columns)

    def add_test_data(self, X: pd.DataFrame) -> "TimeSeriesDataset":
        assert self.time_col in X.columns
        train_data = self.all_data[self.all_data[self.time_col] < X[self.time_col].min()]
        return TimeSeriesDataset(train_data, self.time_col, self.target_names, self.time_idx, X)

    @staticmethod
    def to_dataframe(X, y, target_names: List[str], time_col: str):
        assert len(X) == len(y), "X_val and y_val must have the same length"
        validate_data_basic(X, y)
        # coerce them into a dataframe
        val_df = normalize_ts_data(X, target_names, time_col, y)
        return val_df

    @property
    def all_data(self):
        if len(self.test_data):
            return pd.concat([self.train_data, self.test_data], axis=0)
        else:
            return self.train_data

    @property
    def regressors(self):
        return self.time_varying_known_categoricals + self.time_varying_known_reals

    @property
    def end_date(self):
        test_len = 0 if self.test_data is None else len(self.test_data)
        data = self.test_data if test_len else self.train_data
        return data.iloc[-1][self.time_col]

    def _X(self, df: pd.DataFrame):
        features = [col for col in df.columns if col not in self.target_names]
        return df[features]

    def _y(self, df: pd.DataFrame):
        if len(self.target_names) > 1:
            return df[self.target_names]
        else:
            return df[self.target_names[0]]

    @property
    def X_train(self) -> pd.DataFrame:
        return self._X(self.train_data)

    @property
    def X_val(self) -> pd.DataFrame:
        return self._X(self.test_data)

    @property
    def X_all(self) -> pd.DataFrame:
        return pd.concat([self.X_train, self.X_val], axis=0)

    @property
    def y_train(self) -> pd.DataFrame:
        return self._y(self.train_data)

    @property
    def y_val(self) -> pd.DataFrame:
        return self._y(self.test_data)

    @property
    def y_all(self) -> pd.DataFrame:
        return self._y(self.all_data)

    def next_scale(self) -> int:
        scale_map = {"D": 7, "MS": 12}
        return scale_map.get(self.frequency, 8)

    def known_features_to_floats(self, train: bool, drop_first: bool = True) -> np.ndarray:
        # this is a bit tricky as shapes for train and test data must match, so need to encode together
        combined = pd.concat(
            [
                self.train_data,
                self.test_data,
            ],
            ignore_index=True,
        )

        cat_one_hots = pd.get_dummies(
            combined[self.time_varying_known_categoricals],
            columns=self.time_varying_known_categoricals,
            drop_first=drop_first,
        ).values.astype(float)

        reals = combined[self.time_varying_known_reals].values.astype(float)
        both = np.concatenate([reals, cat_one_hots], axis=1)

        if train:
            return both[: len(self.train_data)]
        else:
            return both[len(self.train_data) :]

    # def unique_dimension_values(self) -> np.ndarray:
    #     # this is the same set for train and test data, by construction
    #     return self.combine_dims(self.train_data).unique()
    #
    # def combine_dims(self, df):
    #     return df.apply(lambda row: tuple([row[d] for d in self.dimensions]), axis=1)

    def to_univariate(self) -> Dict[str, "TimeSeriesDataset"]:
        """
        Convert a multivariate TrainingData  to a dict of univariate ones
        @param df:
        @return:
        """

        train_dims = self.combine_dims(self.train_data)
        test_dims = self.combine_dims(self.test_data)

        out = {}
        for d in train_dims.unique():
            out[d] = copy.copy(self)
            out[d].train_data = self.train_data[train_dims == d]
            out[d].test_data = self.test_data[test_dims == d]
        return out

    def move_validation_boundary(self, steps: int) -> "TimeSeriesDataset":
        out = copy.copy(self)
        if steps > 0:
            out.train_data = pd.concat([self.train_data, self.test_data[:steps]])
            out.test_data = self.test_data[steps:]
        elif steps < 0:
            out.train_data = self.train_data[:steps]
            if len(self.test_data):
                out.test_data = pd.concat([self.train_data[steps:], self.test_data])
            else:
                out.test_data = self.train_data[steps:]

        return out

    def cv_train_val_sets(
        self, n_splits: int, val_length: int, step_size: int
    ) -> Generator["TimeSeriesDataset", None, None]:
        max_index = len(self.train_data) - 1
        for i in range(n_splits):
            out = copy.copy(self)
            val_start = max_index - (n_splits - i - 1) * step_size - val_length
            out.train_data = self.train_data[:val_start]
            out.test_data = self.train_data[val_start : val_start + val_length]
            yield out

    def filter(self, filter_fun: Callable) -> "TimeSeriesDataset":
        if filter_fun is None:
            return self
        out = copy.copy(self)
        out.train_data = self.train_data[filter_fun]
        out.test_data = self.test_data[filter_fun]
        return out

    def prettify_prediction(self, y_pred: Union[pd.DataFrame, pd.Series, np.ndarray]):
        if self.test_data is not None and len(self.test_data):
            assert len(y_pred) == len(self.test_data)

            if isinstance(y_pred, np.ndarray):
                y_pred = pd.DataFrame(data=y_pred, columns=self.target_names, index=self.test_data.index)
            elif isinstance(y_pred, pd.Series):
                assert len(self.target_names) == 1, "Not enough columns in y_pred"
                y_pred.name = self.target_names[0]
                y_pred = pd.DataFrame(y_pred)
                y_pred.index = self.test_data.index
            elif isinstance(y_pred, pd.DataFrame):
                y_pred.index = self.test_data.index

            if self.time_col not in y_pred.columns:
                y_pred[self.time_col] = self.test_data[self.time_col]

        else:
            if isinstance(y_pred, np.ndarray):
                raise ValueError("Can't enrich np.ndarray as self.test_data is None")
            elif isinstance(y_pred, pd.Series):
                assert len(self.target_names) == 1, "Not enough columns in y_pred"
                y_pred = pd.DataFrame({self.target_names[0]: y_pred})
            # TODO auto-create the timestamps for the time column instead of throwing
            raise NotImplementedError("Need a non-None test_data for this to work, for now")

        assert isinstance(y_pred, pd.DataFrame)
        assert self.time_col in y_pred.columns
        assert all([t in y_pred.columns for t in self.target_names])
        return y_pred

    def merge_prediction_with_target(self, y_pred: Union[pd.DataFrame, pd.Series, np.ndarray]):
        y_pred = self.prettify_prediction(y_pred)
        return pd.concat([self.train_data[[self.time_col] + self.target_names], y_pred], axis=0)


def enrich_dataframe(
    df: Union[pd.DataFrame, pd.Series],
    fourier_degree: int,
    remove_constants: bool = False,
    fourier_time: bool = True,
) -> pd.DataFrame:
    if isinstance(df, pd.Series):
        df = pd.DataFrame(df)

    new_cols = []
    for col in df.columns:
        if df[col].dtype.name == "datetime64[ns]":
            extras = monthly_fourier_features(df[col], fourier_degree)
            extras.columns = [f"{col}_{c}" for c in extras.columns]
            extras.index = df.index
            new_cols.append(extras)
            date_feat = date_feature_dict_fourier(df[col]) if fourier_time else date_feature_dict(df[col])
            if remove_constants:
                re_date_feat = {k: v for k, v in date_feat.items() if v.nunique(dropna=False) >= 2}
            else:
                re_date_feat = date_feat

            date_feat = pd.DataFrame(re_date_feat, index=df.index)
            new_cols.append(date_feat)

    return pd.concat([df] + new_cols, axis=1, verify_integrity=True)


def enrich_dataset(
    X: TimeSeriesDataset,
    fourier_degree: int = 0,
    remove_constants: bool = False,
    fourier_time: bool = True,
) -> TimeSeriesDataset:
    new_train = enrich_dataframe(X.train_data, fourier_degree, remove_constants, fourier_time)
    new_test = (
        None if X.test_data is None else enrich_dataframe(X.test_data, fourier_degree, remove_constants, fourier_time)
    )
    return TimeSeriesDataset(
        train_data=new_train,
        time_col=X.time_col,
        target_names=X.target_names,
        time_idx=X.time_idx,
        test_data=new_test,
    )


def date_feature_dict(timestamps: pd.Series) -> dict:
    tmp_dt = timestamps.dt
    column = timestamps.name
    pre_columns_dict = {
        # f"{column}_year": tmp_dt.year, # not stationary
        f"{column}_month": tmp_dt.month,
        # f"{column}_day": tmp_dt.day,# taken care of with monthly fourier features
        f"{column}_hour": tmp_dt.hour,
        f"{column}_minute": tmp_dt.minute,
        f"{column}_second": tmp_dt.second,
        f"{column}_dayofweek": tmp_dt.dayofweek,
        f"{column}_dayofyear": tmp_dt.dayofyear,
        f"{column}_quarter": tmp_dt.quarter,
    }

    new_columns_dict = {}
    for k, v in pre_columns_dict.items():
        new_columns_dict.update(fourier_series(v, k))

    return new_columns_dict


def date_feature_dict_fourier(timestamps: pd.Series) -> dict:
    tmp_dt = timestamps.dt
    column = timestamps.name
    pre_columns_dict = {
        # f"{column}_year": tmp_dt.year, # not stationary
        f"{column}_month": tmp_dt.month / 12.0,
        # f"{column}_day": tmp_dt.day,# taken care of with monthly fourier features
        f"{column}_hour": tmp_dt.hour / 24.0,
        f"{column}_minute": tmp_dt.minute / 60.0,
        f"{column}_second": tmp_dt.second / 60.0,
        f"{column}_dayofweek": tmp_dt.dayofweek / 7.0,
        f"{column}_dayofyear": tmp_dt.dayofyear / 366.0,
        f"{column}_quarter": tmp_dt.quarter / 4.0,
    }

    new_columns_dict = {}
    for k, v in pre_columns_dict.items():
        new_columns_dict.update(fourier_series(v, k))

    return new_columns_dict


def fourier_series(feature: pd.Series, name: str):
    """
    Assume feature goes from 0 to 1 cyclically, transform that into Fourier
    @param feature: input feature
    @return: sin(2pi*feature), cos(2pi*feature)
    """
    return {
        name + "_sin": np.sin(2 * math.pi * feature),
        name + "_cos": np.cos(2 * math.pi * feature),
    }


class DataTransformerTS:
    """Transform input time series training data."""

    def __init__(self, time_col: str, label: Union[str, List[str]], time_idx: str = "time_idx"):
        self.time_col = time_col
        self.time_idx = time_idx
        self.label = label
        self.cat_columns = []
        self.num_columns = []
        self.datetime_columns = []
        self.drop_columns = []

    @property
    def _drop(self):
        return len(self.drop_columns)

    def fit(self, X: Union[DataFrame, np.array], y):
        """Fit transformer.

        Args:
            X: A numpy array or a pandas dataframe of training data.
            y: A numpy array or a pandas series of labels.

        Returns:
            X: Processed numpy array or pandas dataframe of training data.
            y: Processed numpy array or pandas series of labels.
        """
        assert isinstance(X, DataFrame)
        X = X.copy()
        n = X.shape[0]

        assert len(self.num_columns) == 0, "Trying to call fit() twice, something is wrong"

        for column in X.columns:
            # sklearn/utils/validation.py needs int/float values
            if X[column].dtype.name in ("object", "category", "string"):
                if (
                    # drop columns where all values are the same
                    X[column].nunique() == 1
                    # this drops UID-type cols
                    or X[column].nunique(dropna=True) == n - X[column].isnull().sum()
                ):
                    self.drop_columns.append(column)
                elif column != self.time_idx:
                    self.cat_columns.append(column)
            elif X[column].nunique(dropna=True) < 2:
                self.drop_columns.append(column)
            elif X[column].dtype.name in ["datetime64[ns]", "datetime64[s]"]:
                pass  # these will be processed at model level,
                # so they can also be done in the predict method
            else:
                self.num_columns.append(column)

        if self.num_columns:
            self.transformer = ColumnTransformer(
                [
                    (
                        "continuous",
                        SimpleImputer(missing_values=np.nan, strategy="median"),
                        self.num_columns,
                    )
                ]
            )

            self.transformer.fit(X[self.num_columns])
        else:
            self.transformer = None

        # TODO: revisit for multivariate series, and recast for a single df input anyway
        if isinstance(y, Series):
            y = y.rename(self.label)

        if isinstance(y, pd.DataFrame):
            ycol = y[y.columns[0]]
        elif isinstance(y, pd.Series):
            ycol = y
        else:
            raise ValueError("y must be either a pd.Series or a pd.DataFrame at this stage")

        if not pd.api.types.is_numeric_dtype(ycol):
            self.label_transformer = LabelEncoder()
            self.label_transformer.fit(ycol)
        else:
            self.label_transformer = None

    def transform(self, X: Union[DataFrame, np.array], y=None):
        # TODO: revisit for multivariate series, and recast for a single df input anyway
        if self.label_transformer is not None and y is not None:
            if isinstance(y, pd.DataFrame):
                ycol = y[y.columns[0]]
            elif isinstance(y, pd.Series):
                ycol = y
            else:
                raise ValueError("y must be either a pd.Series or a pd.DataFrame at this stage")
            y_tr = self.label_transformer.transform(ycol)
            y.iloc[:] = y_tr.reshape(y.shape)

        X.drop(columns=self.drop_columns, inplace=True)

        for col in self.cat_columns:
            if X[col].dtype.name == "category":
                if "__NAN__" not in X[col].cat.categories:
                    X[col] = X[col].cat.add_categories("__NAN__").fillna("__NAN__")
            else:
                X[col] = X[col].fillna("__NAN__")
                X[col] = X[col].astype("category")

        for column in self.num_columns:
            X[column] = X[column].fillna(np.nan)

        if self.transformer is not None:
            X[self.num_columns] = self.transformer.transform(X[self.num_columns])

        if y is None:
            return X
        return X, y

    def fit_transform(self, X: Union[DataFrame, np.array], y):
        self.fit(X, y)
        return self.transform(X, y)


def create_forward_frame(
    frequency: str,
    steps: int,
    test_end_date: datetime.datetime,
    time_col: str,
):
    start_date = test_end_date + pd.Timedelta(1, frequency)
    times = pd.date_range(
        start=start_date,
        periods=steps,
        freq=frequency,
    )
    return pd.DataFrame({time_col: times})


def normalize_ts_data(X_train_all, target_names, time_col, y_train_all=None):
    if isinstance(X_train_all, TimeSeriesDataset):
        return X_train_all

    if issparse(X_train_all):
        X_train_all = X_train_all.tocsr()

    if isinstance(X_train_all, np.ndarray) and len(X_train_all.shape) == 1:
        X_train_all = np.reshape(X_train_all, (X_train_all.size, 1))

    if isinstance(X_train_all, np.ndarray):
        X_train_all = pd.DataFrame(
            X_train_all,
            columns=[time_col] + [f"x{i}" for i in range(X_train_all.shape[1] - 1)],
        )

    if y_train_all is None:
        return X_train_all
    else:
        if isinstance(y_train_all, np.ndarray):
            # TODO: will need to revisit this when doing multivariate y
            y_train_all = pd.DataFrame(
                y_train_all.reshape(len(X_train_all), -1),
                columns=target_names,
                index=X_train_all.index,
            )
        elif isinstance(y_train_all, pd.Series):
            y_train_all = pd.DataFrame(y_train_all)
            y_train_all.index = X_train_all.index

        dataframe = pd.concat([X_train_all, y_train_all], axis=1)

        return dataframe


def validate_data_basic(X_train_all, y_train_all):
    assert isinstance(X_train_all, np.ndarray) or issparse(X_train_all) or isinstance(X_train_all, pd.DataFrame), (
        "X_train_all must be a numpy array, a pandas dataframe, " "or Scipy sparse matrix."
    )

    assert (
        isinstance(y_train_all, np.ndarray)
        or isinstance(y_train_all, pd.Series)
        or isinstance(y_train_all, pd.DataFrame)
    ), "y_train_all must be a numpy array or a pandas series or DataFrame."

    assert X_train_all.size != 0 and y_train_all.size != 0, "Input data must not be empty, use None if no data"

    assert X_train_all.shape[0] == y_train_all.shape[0], "# rows in X_train must match length of y_train."



import logging
import time
from typing import List

import numpy as np
import pandas as pd
from scipy.sparse import issparse
from sklearn.model_selection import (
    GroupKFold,
    TimeSeriesSplit,
)

from flaml.automl.ml import default_cv_score_agg_func, get_val_loss
from flaml.automl.task.task import (
    TS_FORECAST,
    TS_FORECASTPANEL,
    Task,
    get_classification_objective,
)
from flaml.automl.time_series.ts_data import (
    DataTransformerTS,
    TimeSeriesDataset,
    normalize_ts_data,
)

logger = logging.getLogger(__name__)


class TimeSeriesTask(Task):
    @property
    def estimators(self):
        if self._estimators is None:
            # put this into a function to avoid circular dependency
            from flaml.automl.time_series import (
                ARIMA,
                LGBM_TS,
                RF_TS,
                SARIMAX,
                Average,
                CatBoost_TS,
                ExtraTrees_TS,
                HoltWinters,
                LassoLars_TS,
                Naive,
                Orbit,
                Prophet,
                SeasonalAverage,
                SeasonalNaive,
                TCNEstimator,
                TemporalFusionTransformerEstimator,
                XGBoost_TS,
                XGBoostLimitDepth_TS,
            )

            self._estimators = {
                "xgboost": XGBoost_TS,
                "xgb_limitdepth": XGBoostLimitDepth_TS,
                "rf": RF_TS,
                "lgbm": LGBM_TS,
                "extra_tree": ExtraTrees_TS,
                "arima": ARIMA,
                "sarimax": SARIMAX,
                "holt-winters": HoltWinters,
                "catboost": CatBoost_TS,
                "tft": TemporalFusionTransformerEstimator,
                "lassolars": LassoLars_TS,
                "tcn": TCNEstimator,
                "snaive": SeasonalNaive,
                "naive": Naive,
                "savg": SeasonalAverage,
                "avg": Average,
            }

            if self._estimators["tcn"] is None:
                # remove TCN if import failed
                del self._estimators["tcn"]
                logger.info("Couldn't import pytorch_lightning, skipping TCN estimator")

            try:
                from prophet import Prophet as foo

                self._estimators["prophet"] = Prophet
            except ImportError:
                logger.info("Couldn't import Prophet, skipping")

            try:
                from orbit.models import DLT

                self._estimators["orbit"] = Orbit
            except ImportError:
                logger.info("Couldn't import orbit, skipping")

        return self._estimators

    # processed
    def validate_data(
        self,
        automl,
        state,
        X_train_all,
        y_train_all,
        dataframe,
        label,
        X_val=None,
        y_val=None,
        groups_val=None,
        groups=None,
    ):
        # first beat the data into a TimeSeriesDataset shape
        if isinstance(X_train_all, TimeSeriesDataset):
            # in this case, we're most likely being called by another FLAML instance
            # so all the preliminary cleaning has already been done
            pre_data = X_train_all
            val_len = len(pre_data.X_val)
        else:
            if label is None and dataframe is not None:
                raise ValueError("If data is specified via dataframe parameter, you must also specify label")

            if isinstance(y_train_all, pd.Series):
                label = y_train_all.name
            elif isinstance(y_train_all, np.ndarray):
                label = "y"  # Prophet convention

            if isinstance(label, str):
                target_names = [label]
            else:
                target_names = label

            if self.time_col is None:
                if isinstance(X_train_all, pd.DataFrame):
                    assert dataframe is None, "One of dataframe and X arguments must be None"
                    self.time_col = X_train_all.columns[0]
                elif dataframe is not None:
                    assert X_train_all is None, "One of dataframe and X arguments must be None"
                    self.time_col = dataframe.columns[0]
                else:
                    self.time_col = "ds"

            automl._df = True

            if X_train_all is not None:
                assert y_train_all is not None, "If X_train_all is not None, y_train_all must also be"
                assert dataframe is None, "If X_train_all is provided, dataframe must be None"
                dataframe = TimeSeriesDataset.to_dataframe(X_train_all, y_train_all, target_names, self.time_col)

            elif dataframe is not None:
                assert label is not None, "A label or list of labels must be provided."
                assert isinstance(dataframe, pd.DataFrame), "dataframe must be a pandas DataFrame"
                assert label in dataframe.columns, f"{label} must a column name in dataframe"
            else:
                raise ValueError("Must supply either X_train_all and y_train_all, or dataframe and label")

            try:
                dataframe[self.time_col] = pd.to_datetime(dataframe[self.time_col])
            except Exception:
                raise ValueError(
                    f"For '{TS_FORECAST}' task, time column {self.time_col} must contain timestamp values."
                )

            dataframe = remove_ts_duplicates(dataframe, self.time_col)

            if X_val is not None:
                assert y_val is not None, "If X_val is not None, y_val must also be"
                val_df = TimeSeriesDataset.to_dataframe(X_val, y_val, target_names, self.time_col)
                val_len = len(val_df)
            else:
                val_len = 0
                val_df = None

            pre_data = TimeSeriesDataset(
                train_data=dataframe,
                time_col=self.time_col,
                target_names=target_names,
                test_data=val_df,
            )

        # TODO: should the transformer be a property of the dataset instead?
        automl._transformer = DataTransformerTS(self.time_col, label)
        Xt, yt = automl._transformer.fit_transform(pre_data.X_all, pre_data.y_all)

        df_t = pd.concat([Xt, yt], axis=1)

        data = TimeSeriesDataset(
            train_data=df_t,
            time_col=pre_data.time_col,
            target_names=pre_data.target_names,
        ).move_validation_boundary(-val_len)

        # now setup the properties of all the other relevant objects

        # TODO: where are these used? Replace with pointers to data?
        automl._X_train_all, automl._y_train_all = Xt, yt

        # TODO: where are these used?
        automl._nrow, automl._ndim = data.X_train.shape

        # make a property instead? Or just fix the call?
        automl._label_transformer = automl._transformer.label_transformer

        automl._feature_names_in_ = (
            automl._X_train_all.columns.to_list() if hasattr(automl._X_train_all, "columns") else None
        )

        self.time_col = data.time_col
        self.target_names = data.target_names

        automl._state.X_val = data
        automl._state.X_train = data
        automl._state.y_train = None
        automl._state.y_val = None
        if data.test_data is not None and len(data.test_data) > 0:
            automl._state.X_train_all = data.move_validation_boundary(len(data.test_data))
        else:
            automl._state.X_train_all = data
        automl._state.y_train_all = None

        automl._state.data_size = data.train_data.shape
        automl.data_size_full = len(data.all_data)
        automl._state.groups = None
        automl._sample_weight_full = None

    def prepare_data(
        self,
        state,
        X_train_all,
        y_train_all,
        auto_argument,
        eval_method,
        split_type,
        split_ratio,
        n_splits,
        data_is_df,
        sample_weight_full,
        time_col=None,
    ):
        state.kf = None
        state.data_size_full = len(y_train_all)

        if split_type in ["uniform", "stratified"]:
            raise ValueError(f"Split type {split_type} is not valid for time series")

        state.groups = None
        state.groups_all = None
        state.groups_val = None

        ts_data = state.X_val
        no_test_data = ts_data is None or ts_data.test_data is None or len(ts_data.test_data) == 0
        if no_test_data and eval_method == "holdout":
            # NOTE: _prepare_data is before kwargs is updated to fit_kwargs_by_estimator
            period = state.fit_kwargs["period"]

            if self.name == TS_FORECASTPANEL:
                # TODO: move this into the TimeSeriesDataset class
                X_train_all = ts_data.X_train
                y_train_all = ts_data.y_train

                X_train_all["time_idx"] -= X_train_all["time_idx"].min()
                X_train_all["time_idx"] = X_train_all["time_idx"].astype("int")
                ids = state.fit_kwargs["group_ids"].copy()
                ids.append(ts_data.time_col)
                ids.append("time_idx")
                y_train_all = pd.DataFrame(y_train_all)
                y_train_all[ids] = X_train_all[ids]
                X_train_all = X_train_all.sort_values(ids)
                y_train_all = y_train_all.sort_values(ids)
                training_cutoff = X_train_all["time_idx"].max() - period
                X_train = X_train_all[lambda x: x.time_idx <= training_cutoff]
                y_train = y_train_all[lambda x: x.time_idx <= training_cutoff].drop(columns=ids)
                X_val = X_train_all[lambda x: x.time_idx > training_cutoff]
                y_val = y_train_all[lambda x: x.time_idx > training_cutoff].drop(columns=ids)

                train_data = normalize_ts_data(
                    X_train,
                    ts_data.target_names,
                    ts_data.time_col,
                    y_train,
                )
                test_data = normalize_ts_data(
                    X_val,
                    ts_data.target_names,
                    ts_data.time_col,
                    y_val,
                )
                ts_data = TimeSeriesDataset(
                    train_data,
                    ts_data.time_col,
                    ts_data.target_names,
                    ts_data.frequency,
                    test_data,
                )
                state.X_val = ts_data
                state.X_train = ts_data

            else:
                # if eval_method = holdout, make holdout data
                num_samples = ts_data.train_data.shape[0]
                assert period < num_samples, f"period={period}>#examples={num_samples}"
                state.X_val = ts_data.move_validation_boundary(-period)
                state.X_train = state.X_val

        if eval_method != "holdout":
            if self.name != TS_FORECASTPANEL:
                period = state.fit_kwargs[
                    "period"
                ]  # NOTE: _prepare_data is before kwargs is updated to fit_kwargs_by_estimator
                step_size = state.fit_kwargs.get("cv_step_size", period)

                ts_data = state.X_train
                if n_splits * step_size + 2 * period > ts_data.y_train.size:
                    n_splits = int((ts_data.y_train.size - 2 * period) / step_size)
                    assert n_splits >= 2, (
                        f"cross validation for forecasting period={period}"
                        f" requires input data with at least {2*period + 2*step_size} examples."
                    )
                    logger.info(f"Using nsplits={n_splits} due to data size limit.")
                state.kf = TimeSeriesSplit(n_splits=n_splits, test_size=period)
                state.kf.step_size = step_size

            else:
                n_groups = ts_data.X_train.groupby(state.fit_kwargs.get("group_ids")).ngroups
                period = state.fit_kwargs["period"]
                state.kf = TimeSeriesSplit(n_splits=n_splits, test_size=period * n_groups)

    # TODO: move task detection to Task.__init__!
    def decide_split_type(
        self,
        split_type,
        y_train_all,
        fit_kwargs,
        groups=None,
    ) -> str:
        # TODO: move into task creation!!!
        if self.name == "classification":
            self.name = get_classification_objective(len(np.unique(y_train_all)))

        # TODO: do we need this?
        if not isinstance(split_type, str):
            assert hasattr(split_type, "split") and hasattr(
                split_type, "get_n_splits"
            ), "split_type must be a string or a splitter object with split and get_n_splits methods."
            assert (
                not isinstance(split_type, GroupKFold) or groups is not None
            ), "GroupKFold requires groups to be provided."
            return split_type

        else:
            assert split_type in ["auto", "time"]
            assert isinstance(
                fit_kwargs.get("period"),
                int,  # NOTE: _decide_split_type is before kwargs is updated to fit_kwargs_by_estimator
            ), f"missing a required integer 'period' for '{TS_FORECAST}' task."
            if fit_kwargs.get("group_ids"):
                # TODO (MARK) This will likely not play well with the task class
                self.name = TS_FORECASTPANEL
                assert isinstance(
                    fit_kwargs.get("group_ids"), list
                ), f"missing a required List[str] 'group_ids' for '{TS_FORECASTPANEL}' task."
            return "time"

    # TODO: merge with preprocess() below
    def _preprocess(self, X, transformer=None):
        if isinstance(X, List):
            try:
                if isinstance(X[0], List):
                    X = [x for x in zip(*X)]
                X = pd.DataFrame(
                    dict(
                        [
                            (transformer._str_columns[idx], X[idx])
                            if isinstance(X[0], List)
                            else (transformer._str_columns[idx], [X[idx]])
                            for idx in range(len(X))
                        ]
                    )
                )
            except IndexError:
                raise IndexError("Test data contains more columns than training data, exiting")
        elif isinstance(X, int):
            return X
        elif issparse(X):
            X = X.tocsr()
        if self.is_ts_forecast():
            X = pd.DataFrame(X)
        if transformer:
            X = transformer.transform(X)
        return X

    def preprocess(self, X, transformer=None):
        if isinstance(X, pd.DataFrame) or isinstance(X, np.ndarray) or isinstance(X, pd.Series):
            X = X.copy()
            X = normalize_ts_data(X, self.target_names, self.time_col)
            return self._preprocess(X, transformer)
        elif isinstance(X, int):
            return X
        else:
            raise ValueError(f"unknown type of X, {X.__class__}")

    def evaluate_model_CV(
        self,
        config,
        estimator,
        X_train_all,
        y_train_all,
        budget,
        kf,
        eval_metric,
        best_val_loss,
        cv_score_agg_func=None,
        log_training_metric=False,
        fit_kwargs={},
        free_mem_ratio=0,  # what is this for?
    ):
        if cv_score_agg_func is None:
            cv_score_agg_func = default_cv_score_agg_func
        start_time = time.time()
        val_loss_folds = []
        log_metric_folds = []
        metric = None
        train_time = pred_time = 0
        total_fold_num = 0
        n = kf.get_n_splits()
        if self.is_classification():
            labels = np.unique(y_train_all)
        else:
            labels = fit_kwargs.get("label_list")  # pass the label list on to compute the evaluation metric
        ts_data = X_train_all
        budget_per_train = budget / n
        ts_data = X_train_all
        for data in ts_data.cv_train_val_sets(kf.n_splits, kf.test_size, kf.step_size):
            estimator.cleanup()
            val_loss_i, metric_i, train_time_i, pred_time_i = get_val_loss(
                config,
                estimator,
                X_train=data,
                y_train=None,
                X_val=data,
                y_val=None,
                eval_metric=eval_metric,
                labels=labels,
                budget=budget_per_train,
                log_training_metric=log_training_metric,
                fit_kwargs=fit_kwargs,
                task=self,
                weight_val=None,
                groups_val=None,
                free_mem_ratio=free_mem_ratio,
            )
            if isinstance(metric_i, dict) and "intermediate_results" in metric_i:
                del metric_i["intermediate_results"]
            total_fold_num += 1
            val_loss_folds.append(val_loss_i)
            log_metric_folds.append(metric_i)
            train_time += train_time_i
            pred_time += pred_time_i
            if time.time() - start_time >= budget:
                break
        val_loss, metric = cv_score_agg_func(val_loss_folds, log_metric_folds)
        n = total_fold_num
        pred_time /= n
        return val_loss, metric, train_time, pred_time

    def default_estimator_list(self, estimator_list: List[str], is_spark_dataframe: bool) -> List[str]:
        assert not is_spark_dataframe, "Spark is not yet supported for time series"

        # TODO: why not do this if/then in the calling function?
        if "auto" != estimator_list:
            return estimator_list

        if self.is_ts_forecastpanel():
            return ["tft"]

        estimator_list = [
            "lgbm",
            "rf",
            "xgboost",
            "extra_tree",
            "xgb_limitdepth",
        ]

        # Catboost appears to be way slower than the others, don't include it by default
        # try:
        #     import catboost
        #
        #     estimator_list.append("catboost")
        # except ImportError:
        #     pass

        if self.is_regression():
            estimator_list += ["arima", "sarimax"]

            try:
                import prophet

                estimator_list.append("prophet")
            except ImportError:
                pass

        return estimator_list

    def default_metric(self, metric: str) -> str:
        assert self.is_ts_forecast(), "If this is not a TS forecasting task, this code should never have been called"
        if metric == "auto":
            return "mape"
        else:
            return metric

    @staticmethod
    def prepare_sample_train_data(automlstate, sample_size):
        # we take the tail, rather than the head, for compatibility with time series

        shift = sample_size - len(automlstate.X_train.train_data)
        sampled_X_train = automlstate.X_train.move_validation_boundary(shift)

        return sampled_X_train, None, None, None


def remove_ts_duplicates(
    X,
    time_col,
):
    """
    Assumes the targets are included
    @param X:
    @param time_col:
    @param y:
    @return:
    """

    duplicates = X.duplicated()

    if any(duplicates):
        logger.warning("Duplicate timestamp values found in timestamp column. " f"\n{X.loc[duplicates, X][time_col]}")
        X = X.drop_duplicates()
        logger.warning("Removed duplicate rows based on all columns")
        assert (
            X[[X.columns[0]]].duplicated() is None
        ), "Duplicate timestamp values with different values for other columns."

    return X



import time

try:
    import pandas as pd
    from pandas import DataFrame, Series, to_datetime
except ImportError:

    class PD:
        pass

    pd = PD()
    pd.DataFrame = None
    pd.Series = None
    DataFrame = Series = None

from flaml import tune
from flaml.automl.data import add_time_idx_col
from flaml.automl.time_series.ts_data import TimeSeriesDataset
from flaml.automl.time_series.ts_model import TimeSeriesEstimator


class TemporalFusionTransformerEstimator(TimeSeriesEstimator):
    """The class for tuning Temporal Fusion Transformer"""

    @classmethod
    def search_space(cls, data, task, pred_horizon, **params):
        space = {
            "gradient_clip_val": {
                "domain": tune.loguniform(lower=0.01, upper=100.0),
                "init_value": 0.01,
            },
            "hidden_size": {
                "domain": tune.lograndint(lower=8, upper=512),
                "init_value": 16,
            },
            "hidden_continuous_size": {
                "domain": tune.randint(lower=1, upper=65),
                "init_value": 8,
            },
            "attention_head_size": {
                "domain": tune.randint(lower=1, upper=5),
                "init_value": 4,
            },
            "dropout": {
                "domain": tune.uniform(lower=0.1, upper=0.3),
                "init_value": 0.1,
            },
            "learning_rate": {
                "domain": tune.loguniform(lower=0.00001, upper=1.0),
                "init_value": 0.001,
            },
        }
        return space

    def transform_ds(self, X_train: TimeSeriesDataset, y_train, **kwargs):
        self.data = X_train.train_data

        max_prediction_length = kwargs["period"]
        self.max_encoder_length = kwargs["max_encoder_length"]
        training_cutoff = self.data["time_idx"].max() - max_prediction_length

        from pytorch_forecasting import TimeSeriesDataSet
        from pytorch_forecasting.data import GroupNormalizer

        self.group_ids = kwargs["group_ids"].copy()
        training = TimeSeriesDataSet(
            self.data[lambda x: x.time_idx <= training_cutoff],
            time_idx="time_idx",
            target=X_train.target_names[0],
            group_ids=self.group_ids,
            min_encoder_length=kwargs.get(
                "min_encoder_length", self.max_encoder_length // 2
            ),  # keep encoder length long (as it is in the validation set)
            max_encoder_length=self.max_encoder_length,
            min_prediction_length=1,
            max_prediction_length=max_prediction_length,
            static_categoricals=kwargs.get("static_categoricals", []),
            static_reals=kwargs.get("static_reals", []),
            time_varying_known_categoricals=kwargs.get("time_varying_known_categoricals", []),
            time_varying_known_reals=kwargs.get("time_varying_known_reals", []),
            time_varying_unknown_categoricals=kwargs.get("time_varying_unknown_categoricals", []),
            time_varying_unknown_reals=kwargs.get("time_varying_unknown_reals", []),
            variable_groups=kwargs.get(
                "variable_groups", {}
            ),  # group of categorical variables can be treated as one variable
            lags=kwargs.get("lags", {}),
            target_normalizer=GroupNormalizer(
                groups=kwargs["group_ids"], transformation="softplus"
            ),  # use softplus and normalize by group
            add_relative_time_idx=True,
            add_target_scales=True,
            add_encoder_length=True,
        )

        # create validation set (predict=True) which means to predict the last max_prediction_length points in time
        # for each series
        validation = TimeSeriesDataSet.from_dataset(training, self.data, predict=True, stop_randomization=True)

        # create dataloaders for model
        batch_size = kwargs.get("batch_size", 64)
        train_dataloader = training.to_dataloader(train=True, batch_size=batch_size, num_workers=0)
        val_dataloader = validation.to_dataloader(train=False, batch_size=batch_size * 10, num_workers=0)

        return training, train_dataloader, val_dataloader

    def fit(self, X_train, y_train, budget=None, **kwargs):
        import warnings

        import pytorch_lightning as pl
        import torch
        from pytorch_forecasting import TemporalFusionTransformer
        from pytorch_forecasting.metrics import QuantileLoss
        from pytorch_lightning.callbacks import EarlyStopping, LearningRateMonitor
        from pytorch_lightning.loggers import TensorBoardLogger

        # a bit of monkey patching to fix the MacOS test
        # all the log_prediction method appears to do is plot stuff, which ?breaks github tests
        def log_prediction(*args, **kwargs):
            pass

        TemporalFusionTransformer.log_prediction = log_prediction

        warnings.filterwarnings("ignore")
        current_time = time.time()
        super().fit(X_train, **kwargs)
        training, train_dataloader, val_dataloader = self.transform_ds(X_train, y_train, **kwargs)
        params = self.params.copy()
        gradient_clip_val = params.pop("gradient_clip_val", None)
        params.pop("n_jobs", None)
        max_epochs = kwargs.get("max_epochs", 20)
        early_stop_callback = EarlyStopping(monitor="val_loss", min_delta=1e-4, patience=10, verbose=False, mode="min")
        lr_logger = LearningRateMonitor()  # log the learning rate
        logger = TensorBoardLogger(kwargs.get("log_dir", "lightning_logs"))  # logging results to a tensorboard
        default_trainer_kwargs = dict(
            gpus=self._kwargs.get("gpu_per_trial", [0]) if torch.cuda.is_available() else None,
            max_epochs=max_epochs,
            gradient_clip_val=gradient_clip_val,
            callbacks=[lr_logger, early_stop_callback],
            logger=logger,
        )
        trainer = pl.Trainer(
            **default_trainer_kwargs,
        )
        tft = TemporalFusionTransformer.from_dataset(
            training,
            **params,
            lstm_layers=2,  # 2 is mostly optimal according to documentation
            output_size=7,  # 7 quantiles by default
            loss=QuantileLoss(),
            log_interval=10,  # uncomment for learning rate finder and otherwise, e.g. to 10 for logging every 10 batches
            reduce_on_plateau_patience=4,
        )
        # fit network
        trainer.fit(
            tft,
            train_dataloaders=train_dataloader,
            val_dataloaders=val_dataloader,
        )
        best_model_path = trainer.checkpoint_callback.best_model_path
        best_tft = TemporalFusionTransformer.load_from_checkpoint(best_model_path)
        train_time = time.time() - current_time
        self._model = best_tft
        return train_time

    def predict(self, X):
        ids = self.group_ids.copy()
        ids.append(self.time_col)
        encoder_data = self.data[lambda x: x.time_idx > x.time_idx.max() - self.max_encoder_length]
        # following pytorchforecasting example, make all target values equal to the last data
        last_data_cols = self.group_ids.copy()
        last_data_cols.append(self.target_names[0])
        last_data = self.data[lambda x: x.time_idx == x.time_idx.max()][last_data_cols]
        decoder_data = X.X_val if isinstance(X, TimeSeriesDataset) else X
        if "time_idx" not in decoder_data:
            decoder_data = add_time_idx_col(decoder_data)
        decoder_data["time_idx"] += encoder_data["time_idx"].max() + 1 - decoder_data["time_idx"].min()
        decoder_data = decoder_data.merge(last_data, how="inner", on=self.group_ids)
        decoder_data = decoder_data.sort_values(ids)
        new_prediction_data = pd.concat([encoder_data, decoder_data], ignore_index=True)
        new_prediction_data["time_idx"] = new_prediction_data["time_idx"].astype("int")
        new_raw_predictions = self._model.predict(new_prediction_data)
        index = [decoder_data[idx].to_numpy() for idx in ids]
        predictions = pd.Series(new_raw_predictions.numpy().ravel(), index=index)
        return predictions
